# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402, D412
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional as _Optional
from typing import Sequence, Tuple, TypeVar, Union

from onnx import TypeProto
from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset14 import Opset14
from onnxscript.onnx_types import (
    BFLOAT16,
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset15(Opset14):
    def __new__(cls):
        return Opset.__new__(cls, "", 15)

    T_BatchNormalization = TypeVar("T_BatchNormalization", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    T1_BatchNormalization = TypeVar("T1_BatchNormalization", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    T2_BatchNormalization = TypeVar("T2_BatchNormalization", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    def BatchNormalization(
        self,
        X: T_BatchNormalization,
        scale: T1_BatchNormalization,
        B: T1_BatchNormalization,
        input_mean: T2_BatchNormalization,
        input_var: T2_BatchNormalization,
        *,
        epsilon: float = 9.999999747378752e-06,
        momentum: float = 0.8999999761581421,
        training_mode: int = 0,
    ) -> Tuple[T_BatchNormalization, T2_BatchNormalization, T2_BatchNormalization]:
        r"""[🌐 BatchNormalization(15)](https://onnx.ai/onnx/operators/onnx__BatchNormalization.html#batchnormalization-15 "Online Documentation")


        Carries out batch normalization as described in the paper
        https://arxiv.org/abs/1502.03167. Depending on the mode it is being run,
        There are five required inputs 'X', 'scale', 'B', 'input_mean' and
        'input_var'.
        Note that 'input_mean' and 'input_var' are expected to be the estimated
        statistics in inference mode (training_mode=False, default),
        and the running statistics in training mode (training_mode=True).
        There are multiple cases for the number of outputs, which we list below:

        * Output case #1: Y, running_mean, running_var (training_mode=True)
        * Output case #2: Y (training_mode=False)

        When training_mode=False, extra outputs are invalid.
        The outputs are updated as follows when training_mode=True:
        ::

            running_mean = input_mean * momentum + current_mean * (1 - momentum)
            running_var = input_var * momentum + current_var * (1 - momentum)

            Y = (X - current_mean) / sqrt(current_var + epsilon) * scale + B


        where:
        ::

            current_mean = ReduceMean(X, axis=all_except_channel_index)
            current_var =  ReduceVar(X, axis=all_except_channel_index)


        Notice that `ReduceVar` refers to the population variance, and it equals to
        `sum(sqrd(x_i - x_avg)) / N`
        where `N` is the population size (this formula does not use sample size `N - 1`).

        The computation of ReduceMean and ReduceVar uses float to avoid overflow for float16 inputs.

        When training_mode=False:
        ::

            Y = (X - input_mean) / sqrt(input_var + epsilon) * scale + B



        For previous (depreciated) non-spatial cases, implementors are suggested
        to flatten the input shape to (N x C * D1 * D2 * ... * Dn) before a BatchNormalization Op.
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            X: (differentiable) Input data tensor from the previous operator; dimensions
                are in the form of (N x C x D1 x D2 ... Dn), where N is the batch size,
                C is the number of channels. Statistics are computed for every channel
                of C over N and D1 to Dn dimensions. For image data, input dimensions
                become (N x C x H x W). The op also accepts single dimension input of
                size N in which case C is assumed to be 1

            scale: (differentiable) Scale tensor of shape (C).

            B: (differentiable) Bias tensor of shape (C).

            input_mean: (differentiable) running (training) or estimated (testing) mean
                tensor of shape (C).

            input_var: (differentiable) running (training) or estimated (testing)
                variance tensor of shape (C).

            epsilon: The epsilon value to use to avoid division by zero.

            momentum: Factor used in computing the running mean and variance.e.g.,
                running_mean = running_mean * momentum + mean * (1 - momentum).

            training_mode: If set to true, it indicates BatchNormalization is being used
                for training, and outputs 1 and 2 are to be computed.
        """

        schema = get_schema("BatchNormalization", 15, "")
        op = Op(self, "BatchNormalization", schema)
        return op(
            *self._prepare_inputs(schema, X, scale, B, input_mean, input_var),
            epsilon=epsilon,
            momentum=momentum,
            training_mode=training_mode,
        )

    T1_Bernoulli = TypeVar("T1_Bernoulli", DOUBLE, FLOAT, FLOAT16)

    T2_Bernoulli: TypeAlias = Union[
        BFLOAT16,
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def Bernoulli(
        self,
        input: T1_Bernoulli,
        *,
        dtype: _Optional[int] = None,
        seed: _Optional[float] = None,
    ) -> T2_Bernoulli:
        r"""[🌐 Bernoulli(15)](https://onnx.ai/onnx/operators/onnx__Bernoulli.html#bernoulli-15 "Online Documentation")


        Draws binary random numbers (0 or 1) from a Bernoulli distribution. The input tensor should be a tensor
        containing probabilities p (a value in the range [0,1]) to be used for drawing the binary random number,
        where an output of 1 is produced with probability p and an output of 0 is produced with probability (1-p).

        This operator is non-deterministic and may not produce the same values in different
        implementations (even if a seed is specified).


        Args:
            input: All values in input have to be in the range:[0, 1].

            dtype: The data type for the elements of the output tensor. if not
                specified, we will use the data type of the input tensor.

            seed: (Optional) Seed to the random generator, if not specified we will auto
                generate one.
        """

        schema = get_schema("Bernoulli", 15, "")
        op = Op(self, "Bernoulli", schema)
        return op(*self._prepare_inputs(schema, input), dtype=dtype, seed=seed)

    T1_CastLike = TypeVar(
        "T1_CastLike",
        BFLOAT16,
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_CastLike = TypeVar(
        "T2_CastLike",
        BFLOAT16,
        BOOL,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def CastLike(self, input: T1_CastLike, target_type: T2_CastLike) -> T2_CastLike:
        r"""[🌐 CastLike(15)](https://onnx.ai/onnx/operators/onnx__CastLike.html#castlike-15 "Online Documentation")


        The operator casts the elements of a given input tensor (the first input) to
        the same data type as the elements of the second input tensor.
        See documentation of the Cast operator for further details.


        Args:
            input: (differentiable) Input tensor to be cast.

            target_type: (non-differentiable) The (first) input tensor will be cast to
                produce a tensor of the same type as this (second input) tensor.
        """

        schema = get_schema("CastLike", 15, "")
        op = Op(self, "CastLike", schema)
        return op(*self._prepare_inputs(schema, input, target_type))

    V_Optional = TypeVar(
        "V_Optional",
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    O_Optional: TypeAlias = Union[
        None,
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def Optional(
        self, input: _Optional[V_Optional] = None, *, type: _Optional[TypeProto] = None
    ) -> O_Optional:
        r"""[🌐 Optional(15)](https://onnx.ai/onnx/operators/onnx__Optional.html#optional-15 "Online Documentation")


        Constructs an optional-type value containing either an empty optional of a certain type specified by the attribute,
        or a non-empty value containing the input element.


        Args:
            input: (optional) The input element.

            type: Type of the element in the optional output
        """

        schema = get_schema("Optional", 15, "")
        op = Op(self, "Optional", schema)
        return op(*self._prepare_inputs(schema, input), type=type)

    O_OptionalGetElement = TypeVar(
        "O_OptionalGetElement",
        _Optional[Sequence[BOOL]],
        _Optional[Sequence[COMPLEX128]],
        _Optional[Sequence[COMPLEX64]],
        _Optional[Sequence[DOUBLE]],
        _Optional[Sequence[FLOAT]],
        _Optional[Sequence[FLOAT16]],
        _Optional[Sequence[INT16]],
        _Optional[Sequence[INT32]],
        _Optional[Sequence[INT64]],
        _Optional[Sequence[INT8]],
        _Optional[Sequence[STRING]],
        _Optional[Sequence[UINT16]],
        _Optional[Sequence[UINT32]],
        _Optional[Sequence[UINT64]],
        _Optional[Sequence[UINT8]],
        _Optional[BOOL],
        _Optional[COMPLEX128],
        _Optional[COMPLEX64],
        _Optional[DOUBLE],
        _Optional[FLOAT],
        _Optional[FLOAT16],
        _Optional[INT16],
        _Optional[INT32],
        _Optional[INT64],
        _Optional[INT8],
        _Optional[STRING],
        _Optional[UINT16],
        _Optional[UINT32],
        _Optional[UINT64],
        _Optional[UINT8],
    )

    V_OptionalGetElement: TypeAlias = Union[
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def OptionalGetElement(self, input: O_OptionalGetElement) -> V_OptionalGetElement:
        r"""[🌐 OptionalGetElement(15)](https://onnx.ai/onnx/operators/onnx__OptionalGetElement.html#optionalgetelement-15 "Online Documentation")


        Outputs the element in the optional-type input. It is an error if the input value does not have an element
        and the behavior is undefined in this case.


        Args:
            input: The optional input.
        """

        schema = get_schema("OptionalGetElement", 15, "")
        op = Op(self, "OptionalGetElement", schema)
        return op(*self._prepare_inputs(schema, input))

    O_OptionalHasElement = TypeVar(
        "O_OptionalHasElement",
        _Optional[Sequence[BOOL]],
        _Optional[Sequence[COMPLEX128]],
        _Optional[Sequence[COMPLEX64]],
        _Optional[Sequence[DOUBLE]],
        _Optional[Sequence[FLOAT]],
        _Optional[Sequence[FLOAT16]],
        _Optional[Sequence[INT16]],
        _Optional[Sequence[INT32]],
        _Optional[Sequence[INT64]],
        _Optional[Sequence[INT8]],
        _Optional[Sequence[STRING]],
        _Optional[Sequence[UINT16]],
        _Optional[Sequence[UINT32]],
        _Optional[Sequence[UINT64]],
        _Optional[Sequence[UINT8]],
        _Optional[BOOL],
        _Optional[COMPLEX128],
        _Optional[COMPLEX64],
        _Optional[DOUBLE],
        _Optional[FLOAT],
        _Optional[FLOAT16],
        _Optional[INT16],
        _Optional[INT32],
        _Optional[INT64],
        _Optional[INT8],
        _Optional[STRING],
        _Optional[UINT16],
        _Optional[UINT32],
        _Optional[UINT64],
        _Optional[UINT8],
    )

    B_OptionalHasElement: TypeAlias = BOOL

    def OptionalHasElement(self, input: O_OptionalHasElement) -> B_OptionalHasElement:
        r"""[🌐 OptionalHasElement(15)](https://onnx.ai/onnx/operators/onnx__OptionalHasElement.html#optionalhaselement-15 "Online Documentation")


        Returns true if the optional-type input contains an element. If it is an empty optional-type, this op returns false.


        Args:
            input: The optional input.
        """

        schema = get_schema("OptionalHasElement", 15, "")
        op = Op(self, "OptionalHasElement", schema)
        return op(*self._prepare_inputs(schema, input))

    T_Pow = TypeVar("T_Pow", BFLOAT16, DOUBLE, FLOAT, FLOAT16, INT32, INT64)

    T1_Pow = TypeVar(
        "T1_Pow",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Pow(self, X: T_Pow, Y: T1_Pow) -> T_Pow:
        r"""[🌐 Pow(15)](https://onnx.ai/onnx/operators/onnx__Pow.html#pow-15 "Online Documentation")


        Pow takes input data (Tensor<T>) and exponent Tensor, and
        produces one output data (Tensor<T>) where the function `f(x) = x^exponent`,
        is applied to the data tensor elementwise.
        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        Args:
            X: (differentiable) First operand, base of the exponent.

            Y: (differentiable) Second operand, power of the exponent.
        """

        schema = get_schema("Pow", 15, "")
        op = Op(self, "Pow", schema)
        return op(*self._prepare_inputs(schema, X, Y))

    T_Shape = TypeVar(
        "T_Shape",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_Shape: TypeAlias = INT64

    def Shape(self, data: T_Shape, *, end: _Optional[int] = None, start: int = 0) -> T1_Shape:
        r"""[🌐 Shape(15)](https://onnx.ai/onnx/operators/onnx__Shape.html#shape-15 "Online Documentation")


        Takes a tensor as input and outputs an 1D int64 tensor containing the shape of the input tensor.
        Optional attributes start and end can be used to compute a slice of the input tensor's shape.
        If start axis is omitted, the slice starts from axis 0.
        The end axis, if specified, is exclusive (and the returned value will not include the size of that axis).
        If the end axis is omitted, the axes upto the last one will be included.
        Negative axes indicate counting back from the last axis.
        Note that axes will be clamped to the range [0, r], where r is the
        rank of the input tensor if they are out-of-range (after adding r in the case of
        negative axis). Thus, specifying any end value > r is equivalent to specifying an end
        value of r, and specifying any start value < -r is equivalent to specifying a start
        value of 0. If start > end, the result will be an empty shape.

        Examples:

        ::

            Input tensor with shape: [2, 3, 4]
            No attributes specified.
            Output: [2, 3, 4]



        ::

            Input tensor with shape: [2, 3, 4]
            start: -1
            Output: [4]



        ::

            Input tensor with shape: [2, 3, 4]
            end: -1
            Output: [2, 3]



        ::

            Input tensor with shape: [2, 3, 4]
            start: 1
            end: 2
            Output: [3]




        Args:
            data: (non-differentiable) An input tensor.

            end: (Optional) Ending axis for slicing the shape. Negative value means
                counting dimensions from the back. If omitted, sizes of all axes upto
                (including) the last one will be included.

            start: (Optional) Starting axis for slicing the shape. Default value is
                0.Negative value means counting dimensions from the back.
        """

        schema = get_schema("Shape", 15, "")
        op = Op(self, "Shape", schema)
        return op(*self._prepare_inputs(schema, data), end=end, start=start)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern


class BiasGeluFusion(pattern.RewriteRuleClassBase):
    """Fuses a Bias-Gelu pattern into a single BiasGelu operator.

    Attributes:
        contrib_op (bool): If True, matches the Gelu operator from the 'com.microsoft' domain.
            If False, matches the standard ONNX Gelu operator.
    """

    def __init__(
        self,
        name: str,
        *,
        contrib_op: bool,
    ):
        super().__init__(name)
        self._contrib_op = contrib_op

    def pattern(self, op, input, bias):
        gelu_add = op.Add(input, bias)

        if self._contrib_op:
            return op.Gelu(gelu_add, _domain="com.microsoft", _outputs=["gelu"])
        else:
            return op.Gelu(gelu_add, _outputs=["gelu"])

    def check(self, op, gelu, input, bias, **_) -> pattern.MatchResult:
        check_result = pattern.MatchResult()
        approximate = gelu.producer().attributes.get_string("approximate")
        if approximate is not None and approximate == "tanh":
            return check_result.fail(
                "Gelu operator with 'approximate' set to 'tanh' is not supported."
            )

        if not _ir_utils.has_rank(bias, 1):
            return check_result.fail("bias is not of shape 1D tensor", bias)

        return check_result

    def rewrite(self, op, input, bias, **_):
        return op.BiasGelu(input, bias, _domain="com.microsoft")


bias_gelu_rules = pattern.RewriteRuleSet(
    [
        *BiasGeluFusion.rule("gelu_onnx_op", contrib_op=False).commute(),
        *BiasGeluFusion.rule("gelu_contrib_op", contrib_op=True).commute(),
    ]
)


fuse_bias_gelu = _fusion_utils.apply_fusion_rules(bias_gelu_rules)



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import TypeVar

from onnx.defs import get_schema

from onnxscript.onnx_opset._impl.opset4 import Opset4
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset5(Opset4):
    def __new__(cls):
        return Opset.__new__(cls, "", 5)

    T_Reshape = TypeVar(
        "T_Reshape",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Reshape(self, data: T_Reshape, shape: INT64) -> T_Reshape:
        r"""[🌐 Reshape(5)](https://onnx.ai/onnx/operators/onnx__Reshape.html#reshape-5 "Online Documentation")


        Reshape the input tensor similar to numpy.reshape.
        First input is the data tensor, second input is a shape tensor which specifies the output shape. It outputs the reshaped tensor.
        At most one dimension of the new shape can be -1. In this case, the value is
        inferred from the size of the tensor and the remaining dimensions. A dimension
        could also be 0, in which case the actual dimension value is unchanged (i.e. taken
        from the input tensor). Shape (second input) could be an empty shape, which means converting to a scalar.
        The input tensor's shape and the output tensor's shape are required to have the same number of elements.

        Args:
            data: An input tensor.

            shape: Specified shape for output.
        """

        schema = get_schema("Reshape", 5, "")
        op = Op(self, "Reshape", schema)
        return op(*self._prepare_inputs(schema, data, shape))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import logging
from typing import Sequence, TypeVar, Union

__all__ = [
    "merge_metadata",
    "pattern",
    "rewrite",
    "RewritePass",
    "MatchResult",
    "MatchContext",
    "RewriteRule",
    "RewriteRuleClassBase",
    "RewriteRuleSet",
    "RewriterContext",
    "MatchingTracer",
    "MatchStatus",
    "RULE_NAME_TAG",
]

import onnx
import onnx_ir.passes.common as common_passes

from onnxscript import ir
from onnxscript.rewriter import pattern
from onnxscript.rewriter._basics import MatchContext, MatchingTracer, MatchResult, MatchStatus
from onnxscript.rewriter._rewrite_rule import (
    RULE_NAME_TAG,
    RewriterContext,
    RewriteRule,
    RewriteRuleClassBase,
    RewriteRuleSet,
    merge_metadata,
)
from onnxscript.rewriter.rules.common import (
    _basic_rules,
    _broadcast_to_matmul,
    _cast_constant_of_shape,
    _collapse_slices,
    _fuse_batchnorm,
    _fuse_pad_into_conv,
    _fuse_relus_clips,
    _min_max_to_clip,
    _no_op,
    _redundant_scatter_nd,
    _remove_optional_bias,
)

logger = logging.getLogger(__name__)

_ModelProtoOrIr = TypeVar("_ModelProtoOrIr", onnx.ModelProto, ir.Model)
_DEFAULT_REWRITE_RULES: tuple[pattern.RewriteRule, ...] = (
    *_no_op.rules,  # TODO: merge this rule into constant folding?
    *_broadcast_to_matmul.rules,
    *_cast_constant_of_shape.rules,
    *_collapse_slices.rules,
    *_min_max_to_clip.rules,
    *_fuse_relus_clips.rules,
    *_basic_rules.basic_optimization_rules(),
    *_redundant_scatter_nd.rules,
    *_fuse_pad_into_conv.rules,
    *_fuse_batchnorm.rules,
    *_remove_optional_bias.rules,
)


class RewritePass(ir.passes.InPlacePass):
    def __init__(
        self,
        rules: Sequence[pattern.RewriteRule] | pattern.RewriteRuleSet,
        /,
    ) -> None:
        super().__init__()
        if isinstance(rules, Sequence):
            if not rules:
                raise ValueError("rules must not be empty")
            # Create a pattern rule-set using provided rules
            rules = pattern.RewriteRuleSet(rules)
        assert isinstance(rules, pattern.RewriteRuleSet)
        self.rules: pattern.RewriteRuleSet = rules

    def call(self, model: ir.Model) -> ir.passes.PassResult:
        count = self.rules.apply_to_model(model)
        if count:
            logger.info("Applied %s of general pattern rewrite rules.", count)

        return ir.passes.PassResult(model, bool(count))


def rewrite(
    model: _ModelProtoOrIr,
    pattern_rewrite_rules: Union[Sequence[pattern.RewriteRule], pattern.RewriteRuleSet]
    | None = None,
) -> _ModelProtoOrIr:
    """Rewrite the model using the provided pattern rewrite rules.

    Unused nodes, functions, and opsets will be removed after the rewrite.

    Args:
        model: The model to be rewritten. Can be an ONNX ModelProto or an ir.Model.
        pattern_rewrite_rules: A sequence of pattern rewrite rules or a RewriteRuleSet.
            If not provided, default rules will be applied. If empty, no rules will be applied
            and the original model will be returned.

    Returns:
        The rewritten model as the same type as the input model.
    """
    if pattern_rewrite_rules is None:
        pattern_rewrite_rules = _DEFAULT_REWRITE_RULES
    elif not pattern_rewrite_rules:
        return model

    if isinstance(model, onnx.ModelProto):
        model_ir = ir.serde.deserialize_model(model)
        proto = True
    else:
        model_ir = model
        proto = False

    rewrite_pass = ir.passes.PassManager(
        (
            RewritePass(pattern_rewrite_rules),
            common_passes.RemoveUnusedNodesPass(),
            common_passes.RemoveUnusedFunctionsPass(),
            common_passes.RemoveUnusedOpsetsPass(),
        )
    )
    model_ir = rewrite_pass(model_ir).model
    if proto:
        return ir.serde.serialize_model(model_ir)
    return model_ir  # type: ignore[return-value]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import unittest

import onnxruntime as ort

from onnxscript.backend import onnx_backend


def load_function(obj):
    return ort.InferenceSession(obj.SerializeToString(), providers=("CPUExecutionProvider",))


def run_function(obj, *inputs):
    names = [i.name for i in obj.get_inputs()]
    if len(names) < len(inputs):
        raise AssertionError(f"Got {len(inputs)} inputs but expecting {len(names)}.")
    feeds = {names[i]: inputs[i] for i in range(len(inputs))}
    got = obj.run(None, feeds)
    return got


class TestOnnxBackEnd(unittest.TestCase):
    folder = os.path.join(os.path.abspath(os.path.dirname(__file__)), "onnx_backend_test_code")

    def test_enumerate_onnx_tests(self):
        name = "test_abs"
        code = list(onnx_backend.enumerate_onnx_tests("node", lambda folder: folder == name))
        self.assertEqual(len(code), 1)

    def test_enumerate_onnx_tests_run_one(self):
        done = 0
        for backend_test in onnx_backend.enumerate_onnx_tests(
            "node", lambda folder: folder == "test_abs"
        ):
            self.assertIn(backend_test.name, repr(backend_test))
            self.assertGreater(len(backend_test), 0)
            backend_test.run(load_function, run_function)
            done += 1
        self.assertEqual(done, 1)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, Tuple, TypeVar

from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset2 import Opset2
from onnxscript.onnx_types import DOUBLE, FLOAT, FLOAT16, INT32
from onnxscript.values import Op, Opset


class Opset3(Opset2):
    def __new__(cls):
        return Opset.__new__(cls, "", 3)

    T_GRU = TypeVar("T_GRU", DOUBLE, FLOAT, FLOAT16)

    T1_GRU: TypeAlias = INT32

    def GRU(
        self,
        X: T_GRU,
        W: T_GRU,
        R: T_GRU,
        B: Optional[T_GRU] = None,
        sequence_lens: Optional[T1_GRU] = None,
        initial_h: Optional[T_GRU] = None,
        *,
        activation_alpha: Optional[Sequence[float]] = None,
        activation_beta: Optional[Sequence[float]] = None,
        activations: Optional[Sequence[str]] = None,
        clip: Optional[float] = None,
        direction: str = "forward",
        hidden_size: Optional[int] = None,
        linear_before_reset: int = 0,
        output_sequence: int = 0,
    ) -> Tuple[T_GRU, T_GRU]:
        r"""[🌐 GRU(3)](https://onnx.ai/onnx/operators/onnx__GRU.html#gru-3 "Online Documentation")


        Computes an one-layer GRU. This operator is usually supported via some custom
        implementation such as CuDNN.

        Notations:

        `X` - input tensor

        `z` - update gate

        `r` - reset gate

        `h` - hidden gate

        `t` - time step (t-1 means previous time step)

        `W[zrh]` - W parameter weight matrix for update, reset, and hidden gates

        `R[zrh]` - R recurrence weight matrix for update, reset, and hidden gates

        `Wb[zrh]` - W bias vectors for update, reset, and hidden gates

        `Rb[zrh]` - R bias vectors for update, reset, and hidden gates

        `WB[zrh]` - W parameter weight matrix for backward update, reset, and hidden gates

        `RB[zrh]` - R recurrence weight matrix for backward update, reset, and hidden gates

        `WBb[zrh]` - W bias vectors for backward update, reset, and hidden gates

        `RBb[zrh]` - R bias vectors for backward update, reset, and hidden gates

        `H` - Hidden state

        `num_directions` - 2 if direction == bidirectional else 1

        Activation functions:

          Relu(x)                - max(0, x)

          Tanh(x)                - (1 - e^{-2x})/(1 + e^{-2x})

          Sigmoid(x)             - 1/(1 + e^{-x})

          (NOTE: Below are optional)

          Affine(x)              - alpha*x + beta

          LeakyRelu(x)           - x if x >= 0 else alpha * x

          ThresholdedRelu(x)     - x if x >= alpha else 0

          ScaledTanh(x)          - alpha*Tanh(beta*x)

          HardSigmoid(x)         - min(max(alpha*x + beta, 0), 1)

          Elu(x)                 - x if x >= 0 else alpha*(e^x - 1)

          Softsign(x)            - x/(1 + |x|)

          Softplus(x)            - log(1 + e^x)

        Equations (Default: f=Sigmoid, g=Tanh):

          - zt = f(Xt*(Wz^T) + Ht-1*Rz + Wbz + Rbz)

          - rt = f(Xt*(Wr^T) + Ht-1*Rr + Wbr + Rbr)

          - ht = g(Xt*(Wh^T) + (rt (.) Ht-1)*Rh + Rbh + Wbh) # default, when linear_before_reset = 0

          - ht = g(Xt*(Wh^T) + (rt (.) (Ht-1*Rh + Rbh) + Wbh) # when linear_before_reset != 0

          - Ht = (1 - zt) (.) ht + zt (.) Ht-1


        Args:
            X: The input sequences packed (and potentially padded) into one 3-D tensor
                with the shape of `[seq_length, batch_size, input_size]`.

            W: The weight tensor for the gates. Concatenation of `W[zrh]` and `WB[zrh]`
                (if bidirectional) along dimension 0. This tensor has shape
                `[num_directions, 3*hidden_size, input_size]`.

            R: The recurrence weight tensor. Concatenation of `R[zrh]` and `RB[zrh]` (if
                bidirectional) along dimension 0. This tensor has shape
                `[num_directions, 3*hidden_size, hidden_size]`.

            B: (optional) The bias tensor for the gates. Concatenation of `[Wb[zrh],
                Rb[zrh]]` and `[WBb[zrh], RBb[zrh]]` (if bidirectional) along dimension
                0. This tensor has shape `[num_directions, 6*hidden_size]`. Optional: If
                not specified - assumed to be 0

            sequence_lens: (optional) Optional tensor specifying lengths of the
                sequences in a batch. If not specified - assumed all sequences in the
                batch to have length `seq_length`. It has shape `[batch_size]`.

            initial_h: (optional) Optional initial value of the hidden. If not specified
                - assumed to be 0. It has shape `[num_directions, batch_size,
                hidden_size]`.

            activation_alpha: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.For example with LeakyRelu, the default
                alpha is 0.01.

            activation_beta: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.

            activations: A list of 2 (or 4 if bidirectional) activation functions for
                update, reset, and hidden gates. The activation functions must be one of
                the activation functions specified above. Optional: See the equations
                for default if not specified.

            clip: Cell clip threshold. Clipping bounds the elements of a tensor in the
                range of [-threshold, +threshold] and is applied to the input of
                activations. No clip if not specified.

            direction: Specify if the RNN is forward, reverse, or bidirectional. Must be
                one of forward (default), reverse, or bidirectional.

            hidden_size: Number of neurons in the hidden layer

            linear_before_reset: When computing the output of the hidden gate, apply the
                linear transformation before multiplying by the output of the reset
                gate.

            output_sequence: The sequence output for the hidden is optional if 0.
                Default 0.
        """

        schema = get_schema("GRU", 3, "")
        op = Op(self, "GRU", schema)
        return op(
            *self._prepare_inputs(schema, X, W, R, B, sequence_lens, initial_h),
            activation_alpha=activation_alpha,
            activation_beta=activation_beta,
            activations=activations,
            clip=clip,
            direction=direction,
            hidden_size=hidden_size,
            linear_before_reset=linear_before_reset,
            output_sequence=output_sequence,
        )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Sequence, Union

import numpy as np
import onnx_ir as ir

import onnxscript.rewriter._fusion_utils as _fusion_utils
from onnxscript.rewriter import _basics, _ir_utils, pattern

"""
GroupQueryAttention: This generalizes MHA by allowing the number of heads to be different
for query and key/value.

We use the following abbreviations for the dimensions:
B: Batch size
S: Sequence length (for current query/key/value)

Hkv: number of heads for key/value
G = number of groups
H: number of heads = G * Hkv

Dh: head size or embedding dimension per head
D: input embedding dimension (hidden size) = H * Dh
Dkv: key/value hidden size = Hkv * Dh

T: total sequence length (after concatenation of past and current key/value)
"""

Dim = Union[int, ir.SymbolicDim]


def _is_model_input(value: ir.Value, name: str, model: ir.Model) -> bool:
    return value in model.graph.inputs and value.name == name


def _causal_mask(
    op,
    input_ids,
    past_kv_cache,
    shape_B111,
    min_val,
    window_size,
    dtype,
):
    """Defines a pattern for a pure causal mask, with optional sliding window support."""
    seq_len = op.Shape(input_ids, end=2, start=1)
    seq_len_0D = op.Squeeze(seq_len)

    past_seq_len = op.Shape(past_kv_cache, end=3, start=2)
    past_seq_len_0D = op.Squeeze(past_seq_len)

    total_seq_len_0D = op.Add(past_seq_len_0D, seq_len_0D)
    total_seq_len = op.Reshape(total_seq_len_0D, [-1])

    current_range = op.Range(past_seq_len_0D, total_seq_len_0D, 1)
    mask_shape = op.Concat(seq_len, total_seq_len, axis=0)
    mask_all_min_expand = op.Expand(min_val, mask_shape)
    # The following Trilu is optional: not used in Phi models, but used in LLama.
    mask_all_min_trilu = op.Trilu(mask_all_min_expand, 1, upper=1)
    mask_all_min = pattern.OrValue([mask_all_min_expand, mask_all_min_trilu])
    total_range_as_row = op.Range(0, total_seq_len_0D, 1)
    current_range_as_column = op.Reshape(current_range, [-1, 1])

    non_causal = op.Greater(total_range_as_row, current_range_as_column)

    # sliding window support:
    current_range_minus_window = op.Sub(current_range_as_column, window_size)
    out_of_sliding_window = op.LessOrEqual(total_range_as_row, current_range_minus_window)
    non_causal_sliding_window = op.Or(non_causal, out_of_sliding_window)

    boolean_mask = pattern.OrValue([non_causal, non_causal_sliding_window])

    float_0_1_mask = op.Cast(boolean_mask, to=dtype)
    float_0_min_mask = op.Mul(mask_all_min, float_0_1_mask)
    mask_4d_11ST = op.Unsqueeze(float_0_min_mask, [0, 1])
    mask_4d_B1ST = op.Expand(mask_4d_11ST, shape_B111)

    return mask_4d_B1ST


class _CausalMaskPattern(pattern.PatternBase):
    def pattern(
        self,
        op,
        input_ids,
        past_kv_cache,
        shape_B111,
        min_val,
        window_size,
        dtype1,
        attn_mask_2d,
        dtype2,
    ):
        causal_mask = _causal_mask(
            op,
            input_ids,
            past_kv_cache,
            shape_B111,
            min_val,
            window_size,
            dtype1,
        )

        attn_mask_4d = op.Unsqueeze(attn_mask_2d, [1, 2])
        attn_mask_4d_cast = op.Cast(attn_mask_4d, to=dtype2)

        sum = op.Add(causal_mask, attn_mask_4d_cast)
        sum_fp32 = op.Cast(sum, to=ir.DataType.FLOAT)
        # The cast is optional, and may be absent if the sum is already in float32.
        sum_fp32 = pattern.OrValue([sum_fp32, sum])
        is_zero = op.Equal(sum_fp32, 0.0)
        result = op.Where(is_zero, min_val, causal_mask)
        return result

    def check(self, context, dtype1, dtype2, min_val, attn_mask_2d, sliding_window=None, **_):
        # Check that attn_mask_2d is the model input "attention_mask"
        if not _is_model_input(attn_mask_2d, "attention_mask", context.model):
            return pattern.MatchResult().fail("Invalid attention_mask input", attn_mask_2d)

        if dtype1.as_int() != dtype2.as_int():
            return pattern.MatchResult().fail("Dtype mismatch", [dtype1, dtype2])

        # Check that min_val is a constant and matches the expected minimum value for the dtype.
        min_value = _ir_utils.get_singleton_value(min_val)
        if min_value is None:
            return pattern.MatchResult().fail("Minval is not a constant.", min_val)
        expected_min_value = np.finfo(min_val.dtype.numpy()).min
        if min_value != expected_min_value:
            return pattern.MatchResult().fail(
                f"Expected min value {expected_min_value}, got {min_value}", min_val
            )

        # TODO(rama) Sliding window: not yet supported.
        if sliding_window:
            return pattern.MatchResult().fail(
                "Sliding window not yet supported", sliding_window
            )
        return True


_causal_mask_pattern = _CausalMaskPattern()


class GroupQueryAttention(pattern.RewriteRuleClassBase):
    def __init__(self):
        super().__init__("GQA", remove_nodes=False)

    def pattern(
        self,
        op,
        query_BSD,
        key_BSDkv,
        value_BSDkv,
        past_key,
        past_value,
        position_ids,
        cos,
        sin,
        mask,
    ):
        # Reshape query from (B, S, D) to (B, S, H, D/H)
        query_BSHDh = op.Reshape(query_BSD, pattern.ANY_VALUE, _outputs=["query_BSHDh"])
        # Qwen variant uses normalization of query/key before rotary embedding:
        # The normalization can happen before (eg., Qwen) or after the Transpose (eg., Gemma).
        query_BSHDh_normalized = op.SimplifiedLayerNormalization(
            query_BSHDh, pattern.ANY_VALUE, axis=-1, _outputs=["query_BSHDh_normalized"]
        )
        query_BSHDh = pattern.OrValue([query_BSHDh, query_BSHDh_normalized])

        # Transpose from (B, S, H, D/H) to (B, H, S, D/H)
        query_BHSDh = op.Transpose(query_BSHDh, perm=[0, 2, 1, 3])

        # Gemma variant uses normalization of query/key before rotary embedding:
        query_BHSDh_normalized = op.SimplifiedLayerNormalization(
            query_BHSDh, pattern.ANY_VALUE, axis=-1, _outputs=["query_BHSDh_normalized"]
        )
        query_BHSDh = pattern.OrValue([query_BHSDh, query_BHSDh_normalized])

        # Reshape key from (B, S, Dkv) to (B, S, Hkv, D/H)
        key_BSHkvDh = op.Reshape(key_BSDkv, pattern.ANY_VALUE, _outputs=["key_BSHkvDh"])
        key_BSHkvDh_normalized = op.SimplifiedLayerNormalization(
            key_BSHkvDh, pattern.ANY_VALUE, axis=-1, _outputs=["key_BSHkvDh_normalized"]
        )
        key_BSHkvDh = pattern.OrValue([key_BSHkvDh, key_BSHkvDh_normalized])

        # Transpose from (B, S, Hkv, D/H) to (B, Hkv, S, D/H)
        key_BHkvSDh = op.Transpose(key_BSHkvDh, perm=[0, 2, 1, 3])

        # Gemma variant uses normalization of query/key before rotary embedding:
        key_BHkvSDh_normalized = op.SimplifiedLayerNormalization(
            key_BHkvSDh, pattern.ANY_VALUE, axis=-1, _outputs=["key_BHkvSDh_normalized"]
        )
        key_BHkvSDh = pattern.OrValue([key_BHkvSDh, key_BHkvSDh_normalized])

        # Reshape value from (B, S, Dkv) to (B, S, Hkv, D/H)
        value_BSHkvDh = op.Reshape(value_BSDkv, pattern.ANY_VALUE, _outputs=["value_BSHkvDh"])
        # Transpose from (B, S, Hkv, D/H) to (B, Hkv, S, D/H)
        value_BHkvSDh = op.Transpose(value_BSHkvDh, perm=[0, 2, 1, 3])

        query_BHSDh_rope = op.RotaryEmbedding(
            query_BHSDh,
            position_ids,
            cos,
            sin,
            _domain="com.microsoft",
            _outputs=["query_BHSDh_rope"],
        )
        key_BHkvSDh_rope = op.RotaryEmbedding(
            key_BHkvSDh,
            position_ids,
            cos,
            sin,
            _domain="com.microsoft",
            _outputs=["key_BHkvSDh_rope"],
        )

        # Concatenate past_key cache and current key, expand across heads
        # that share key/value.

        key_seq_BHkvTDh = op.Concat(past_key, key_BHkvSDh_rope, axis=-2)
        # Concat with past_key is optional:
        key_seq_BHkvTDh = pattern.OrValue([key_seq_BHkvTDh, key_BHkvSDh_rope])
        key_seq_BHkv1TDh = op.Unsqueeze(key_seq_BHkvTDh, [2])
        key_seq_BHkvGTDh = op.Expand(key_seq_BHkv1TDh, pattern.ANY_VALUE)
        key_seq_BHTDh = op.Reshape(
            key_seq_BHkvGTDh, pattern.ANY_VALUE, _outputs=["key_seq_BHTDh"]
        )

        # Concatenate past_value cache and current value, expand across heads
        # that share key/value.
        value_seq_BHkvTDh = op.Concat(past_value, value_BHkvSDh, axis=-2)
        # Concat with past_value is optional:
        value_seq_BHkvTDh = pattern.OrValue([value_seq_BHkvTDh, value_BHkvSDh])
        value_seq_BHkv1TDh = op.Unsqueeze(value_seq_BHkvTDh, [2])
        value_seq_BHkvGTDh = op.Expand(value_seq_BHkv1TDh, pattern.ANY_VALUE)
        value_seq_BHTDh = op.Reshape(
            value_seq_BHkvGTDh, pattern.ANY_VALUE, _outputs=["value_seq_BHTDh"]
        )

        attention_BHSDh = op.SDPA(
            query_BHSDh_rope,
            key_seq_BHTDh,
            value_seq_BHTDh,
            mask,
            key_format="BHSd",
            _domain="ai.onnxruntime._fusion",
        )

        # Transpose attention back to (B, S, H, D/H)
        attention_BSHDh = op.Transpose(attention_BHSDh, perm=[0, 2, 1, 3])
        # Reshape back to (B, S, D)
        attention_BSD = op.Reshape(
            attention_BSHDh, pattern.ANY_VALUE, _outputs=["attention_BSD"]
        )
        return attention_BSD, key_seq_BHkvTDh, value_seq_BHkvTDh

    def check(
        self,
        context: _basics.MatchContext,
        query_BSD,
        key_BSDkv,
        value_BSDkv,
        past_key,
        past_value,
        query_BHSDh_rope,
        key_BHkvSDh_rope,
        query_BSHDh,
        key_BSHkvDh,
        mask,
        query_BSHDh_normalized=None,
        query_BHSDh_normalized=None,
        key_BSHkvDh_normalized=None,
        key_BHkvSDh_normalized=None,
        **_,
    ):
        result = pattern.MatchResult()
        if query_BSHDh_normalized is not None and query_BHSDh_normalized is not None:
            return result.fail(
                "Query normalized twice",
                [query_BSHDh_normalized, query_BHSDh_normalized],
            )
        if key_BSHkvDh_normalized is not None and key_BHkvSDh_normalized is not None:
            return result.fail(
                "Key normalized twice",
                [key_BSHkvDh_normalized, key_BHkvSDh_normalized],
            )
        bindings: dict[str, Dim] = {}

        def no_match(val: ir.Value, dims: Sequence[str]) -> bool:
            return not _fusion_utils.check_shape_bool(bindings, val, dims)

        if no_match(query_BSD, ["B", "S", "D"]):
            return False
        if no_match(key_BSDkv, ["B", "S", "Dkv"]):
            return False
        if no_match(value_BSDkv, ["B", "S", "Dkv"]):
            return False

        if past_key is not None and no_match(past_key, ["B", "Hkv", "P", "Dh"]):
            return False
        if past_value is not None and no_match(past_value, ["B", "Hkv", "P", "Dv"]):
            return False

        # TODO: verify Reshapes:
        # eg.: verify bindings["B"] * bindings["H"] == bindings["B*H"]:
        # and bindings["H"] * bindings["Dh"] == bindings["H*Dh"]:
        # or check Reshape's shape-input value

        num_heads = _ir_utils.get_dim(query_BSHDh, 2)
        kv_num_heads = _ir_utils.get_dim(key_BSHkvDh, 2)
        if not isinstance(num_heads, int):
            return result.fail("Unable to determine num_heads value", query_BSHDh)
        if not isinstance(kv_num_heads, int):
            return result.fail("Unable to determine kv_num_heads value", key_BSHkvDh)
        self.num_heads = num_heads
        self.kv_num_heads = kv_num_heads

        # Rotary embedding attributes
        query_rotary_attributes = query_BHSDh_rope.producer().attributes
        key_rotary_attributes = key_BHkvSDh_rope.producer().attributes
        query_interleaved = query_rotary_attributes.get_int("interleaved", 0)
        key_interleaved = key_rotary_attributes.get_int("interleaved", 0)
        if query_interleaved != key_interleaved:
            return pattern.MatchResult().fail(
                "Rotary embedding interleaved attribute mismatch",
                [query_BHSDh_rope.producer(), key_BHkvSDh_rope.producer()],
            )
        self._interleaved = query_interleaved

        # Check mask:
        mask_node = mask.producer()
        if mask_node is None:
            return pattern.MatchResult().fail("Unhandled mask pattern", mask)
        mask_match_result = _causal_mask_pattern.match(
            context.model,
            context.graph_or_function,
            mask_node,
            check_nodes_are_removable=False,
        )
        if mask_match_result is None:
            return pattern.MatchResult().fail("Mask does not match causal mask pattern", mask)
        # TODO: handle sliding window support in mask

        return True

    def rewrite(
        self,
        op,
        query_BSD,
        key_BSDkv,
        value_BSDkv,
        past_key,
        past_value,
        position_ids,
        cos,
        sin,
        mask,
        query_BSHDh,
        key_BSHkvDh,
        query_BSHDh_normalized=None,
        query_BHSDh_normalized=None,
        key_BSHkvDh_normalized=None,
        key_BHkvSDh_normalized=None,
        **_,
    ):
        # Note that the following optimization is specific to current ORT GenAI attention-mask
        # usage. Specifically, it assumes that the model-input "attention_mask" is a 2D
        # mask with shape [batch_size, sequence_length], and that the mask is a 0/1 mask
        # that is used only to indicate the current tokens. Hence, the input attention_mask
        # is redundant as long as past-sequence-length and current-sequence-length can be
        # computed.

        # Construct seqlens_k and total_seq_length_int32 from position_ids
        # seqlens_k : int32[batch_size] indicates total_sequence-length-1 for each batch
        # position_ids: int64[batch_size, sequence_length] indicates the position of each token
        one_int32_0d = op.Constant(value=ir.tensor(1, dtype=ir.DataType.INT32))
        one_int64_1d = op.Constant(value=ir.tensor([1], dtype=ir.DataType.INT64))
        zero_int64_1d = op.Constant(value=ir.tensor([0], dtype=ir.DataType.INT64))
        seqlens_k_int64 = op.ReduceMax(position_ids, one_int64_1d, keepdims=0)
        seqlens_k = op.Cast(seqlens_k_int64, to=ir.DataType.INT32)
        max_seq_length = op.ReduceMax(seqlens_k, zero_int64_1d, keepdims=0)
        total_seq_length_int32 = op.Add(max_seq_length, one_int32_0d)

        normalized_query = query_BHSDh_normalized or query_BSHDh_normalized
        if normalized_query is not None:
            # We apply normalization without the transpose, which is fused into GQA
            norm_node = normalized_query.producer()
            norm_attrs = norm_node.attributes
            norm_scale = norm_node.inputs[1]
            query_BSHDh_normalized = op.SimplifiedLayerNormalization(
                query_BSHDh, norm_scale, **norm_attrs
            )
            reshape_BSHDh_to_BSD = op.Constant(value_ints=[0, 0, -1])
            query_BSD = op.Reshape(query_BSHDh_normalized, reshape_BSHDh_to_BSD)

        normalized_key = key_BHkvSDh_normalized or key_BSHkvDh_normalized
        if normalized_key is not None:
            # We apply normalization without the transpose, which is fused into GQA
            norm_node = normalized_key.producer()
            norm_attrs = norm_node.attributes
            norm_scale = norm_node.inputs[1]
            key_BSHkvDh_normalized = op.SimplifiedLayerNormalization(
                key_BSHkvDh, norm_scale, **norm_attrs
            )
            reshape_BSHkvDh_to_BSDkv = op.Constant(value_ints=[0, 0, -1])
            key_BSDkv = op.Reshape(key_BSHkvDh_normalized, reshape_BSHkvDh_to_BSDkv)

        return op.GroupQueryAttention(
            query_BSD,
            key_BSDkv,
            value_BSDkv,
            past_key,
            past_value,
            seqlens_k,
            total_seq_length_int32,
            cos,
            sin,
            num_heads=self.num_heads,
            kv_num_heads=self.kv_num_heads,
            do_rotary=1,
            rotary_interleaved=self._interleaved,
            # skipped optional attributes: local_window_size, scale, smooth_softmax, softcap
            _domain="com.microsoft",
            _outputs=3,
        )


_basic_gqa_rule = GroupQueryAttention.rule()

gqa_rules = pattern.RewriteRuleSet([_basic_gqa_rule])

fuse_gqa = _fusion_utils.apply_fusion_rules(gqa_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Deprecated. This module is kept for backward compatibility."""

from __future__ import annotations

from typing import Sequence

import onnx

from onnxscript.rewriter import pattern
from onnxscript.rewriter import rewrite as _rewrite
from onnxscript.rewriter.ort_fusions import ORT_PATTERN_REWRITE_RULES

__all__ = [
    "rewrite",
    "ORT_PATTERN_REWRITE_RULES",
]


def rewrite(
    model_proto: onnx.ModelProto,
    /,
    pattern_rules: Sequence[pattern.RewriteRule] | None = None,
) -> onnx.ModelProto:
    """Rewrite the model using the given rules.

    Args:
        model_proto: The model to rewrite.
        pattern_rules: The pattern rewrite rules to apply. If None, the default rules
            for onnxruntime are used.

    Returns:
        The rewritten model.
    """
    pattern_rules = pattern_rules or ORT_PATTERN_REWRITE_RULES
    return _rewrite(model_proto, pattern_rewrite_rules=pattern_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import numpy as np
import onnx
import parameterized

import onnxscript.optimizer as optimizer
from onnxscript import ir
from onnxscript.optimizer import _constant_folding


class FoldConstantsTest(unittest.TestCase):
    def _fold(
        self,
        model: ir.Model | str,
        onnx_shape_inference: bool = False,
        dce: bool = True,
        **kwargs,
    ):
        if isinstance(model, str):
            model = ir.from_onnx_text(model)
        _constant_folding.fold_constants(
            model, onnx_shape_inference=onnx_shape_inference, **kwargs
        )
        if dce:
            optimizer.remove_unused_nodes(model)
        # Ensure the model is valid after optimization
        onnx.checker.check_model(ir.serde.serialize_model(model))
        return model

    def test_fold_add(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z) {
                two = Constant <value_float=2.0> ()
                four = Add(two, two)
                z = Mul(x, four)
            }
        """

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertIn("four", optimized.graph.initializers)

    def test_fold_cast_like(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z) {
                two = Constant <value_int=2> ()
                two_float = CastLike(two, x)
                four = Add(two_float, two_float)
                z = Mul(x, four)
            }
        """

        optimized = self._fold(model, dce=False)
        self.assertIn("four", optimized.graph.initializers)
        np.testing.assert_equal(
            optimized.graph.initializers["four"].const_value, np.array(4.0)
        )
        # Intermediates should be removed
        self.assertNotIn("two_float", optimized.graph.initializers)

        optimized = self._fold(model, dce=True)
        self.assertEqual(len(optimized.graph), 1)

    def test_fold_shape(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[16, 16] x) => (float[16, 16] z) {
                shape = Shape(x)
                rank = Size(shape)
                two_float = CastLike(rank, x)
                four = Add(two_float, two_float)
                z = Mul(x, four)
            }
        """

        optimized = self._fold(model, dce=False)
        self.assertIn("four", optimized.graph.initializers)
        np.testing.assert_equal(
            optimized.graph.initializers["four"].const_value, np.array(4.0)
        )
        # Intermediates should be removed
        self.assertNotIn("two_float", optimized.graph.initializers)
        self.assertNotIn("rank", optimized.graph.initializers)
        self.assertNotIn("shape", optimized.graph.initializers)

        optimized = self._fold(model, dce=True)
        self.assertEqual(len(optimized.graph), 1)

    def test_fold_shape_slice(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[M, N, 16, 16] x) => (float[M, N, 16, 16] z) {
                shape = Shape <start:int = 2>(x)
                two = Size(shape)
                two_float = CastLike(two, x)
                four = Add(two_float, two_float)
                z = Mul(x, four)
            }
        """

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertIn("four", optimized.graph.initializers)

    def test_fold_if_cond(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[16, 16] x) => (float[16, 16] z) {
                shape = Shape(x)
                rank = Size(shape)
                zero = Constant <value_int=0> ()
                zero_cast = CastLike (zero, rank)
                is_scalar = Equal(zero_cast, rank)
                z = If (is_scalar) <
                    then_branch = then_graph () => (then_z) { then_z = Add (x, x) },
                    else_branch = else_graph () => (else_z) { else_z = Mul (x, x) }
                >
            }
        """

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual(optimized.graph[0].outputs[0].name, "z")
        self.assertEqual(optimized.graph[0].op_type, "Mul")

    def test_fold_inside_if_branch(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[16, 16] x, bool cond) => (float[16, 16] z) {
                two = Constant <value_float=2.0> ()
                z = If (cond) <
                    then_branch = then_graph () => (then_z) {
                        three = Constant <value_float=3.0> ()
                        temp = Add (two, three)
                        then_z = Mul (temp, x)
                    },
                    else_branch = else_graph () => (else_z) {
                        four = Constant <value_float=4.0> ()
                        temp = Add (two, four)
                        else_z = Mul (temp, x)
                    }
                >
            }
        """

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        then_graph = optimized.graph[0].attributes["then_branch"].as_graph()
        self.assertEqual(len(then_graph), 1)
        self.assertIn("temp", then_graph.initializers)
        else_graph = optimized.graph[0].attributes["else_branch"].as_graph()
        self.assertEqual(len(else_graph), 1)
        self.assertIn("temp", else_graph.initializers)

    def test_fold_if_propagate(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[16, 16] x) => (float[16, 16] z) {
                shape = Shape(x)
                rank = Size(shape)
                zero = Constant <value_int=0> ()
                two = Constant <value_float=2.0> ()
                zero_cast = CastLike (zero, rank)
                is_scalar = Equal(zero_cast, rank)
                m = If (is_scalar) <
                    then_branch = then_graph () => (then_z) { then_z = Add (x, x) },
                    else_branch = else_graph () => (else_z) { else_z = Mul (two, two) }
                >
                m_square = Mul (m, m)
                z = Mul (x, m_square)
            }
        """

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertIn("m_square", optimized.graph.initializers)

    def test_fold_redundant_cast(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z) {
                two = Constant <value_float: float = 2.0> ()
                x_cast = CastLike(x, two)
                z = Mul(x_cast, two)
            }
        """

        optimized = self._fold(model, onnx_shape_inference=True)
        self.assertEqual(len(optimized.graph), 2)

    def test_fold_redundant_cast2(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z) {
                two = Constant <value_float: float = 2.0> ()
                z = CastLike(x, two)
            }
        """

        optimized = self._fold(model, onnx_shape_inference=True)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual(optimized.graph[0].op_type, "Identity")
        self.assertEqual(optimized.graph[0].outputs[0].name, "z")
        self.assertEqual(optimized.graph[0].inputs[0].name, "x")

    def test_shape_inference(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (int64[64] x) => (int64[N] z) {
                one = Constant <value_int=1> ()
                cond = Equal(one, one)
                temp = If (cond) <
                    then_branch = then_graph () => (then_z) {
                        shape1 = Constant <value_ints=[8, 8]> ()
                        then_z = Reshape(x, shape1)
                    },
                    else_branch = else_graph () => (else_z) {
                        shape2 = Constant <value_ints=[64]> ()
                        else_z = Reshape(x, shape2)
                    }>
                shape = Shape(temp)   # shape = [8, 8] or [64], but [8, 8] after constant propagation
                rank = Size(shape)    # rank = 2 or 1, but 2 after constant propagation
                C = Add (rank, rank)
                z = Mul(x, C)
            }
        """

        optimized = self._fold(model, onnx_shape_inference=True)
        self.assertEqual(len(optimized.graph), 1)
        self.assertIn("C", optimized.graph.initializers)

    def test_static_split_to_sequence_with_scalar_split_and_squence_at_is_folded_as_split(
        self,
    ):
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[1,512] x) => (float[1,512] return_val) {
   int64_128 = Constant <value: tensor = int64 int64_128 {128}> ()
   splits = SplitToSequence <axis: int = 1> (x, int64_128)
   int64_0 = Constant <value: tensor = int64 int64_0 {0}> ()
   split_0 = SequenceAt (splits, int64_0)
   int64_1 = Constant <value: tensor = int64 int64_1 {1}> ()
   split_1 = SequenceAt (splits, int64_1)
   int64_2 = Constant <value: tensor = int64 int64_2 {2}> ()
   split_2 = SequenceAt (splits, int64_2)
   int64_3 = Constant <value: tensor = int64 int64_3 {3}> ()
   split_3 = SequenceAt (splits, int64_3)
   return_val = Concat <axis: int = 1> (split_0, split_1, split_2, split_3)
}"""

        # TODO: There is an unrelated limitation that `symbolic_value` is not
        # utilized when the value is only referenced by graph output.
        # E.g., the following test model will not have this optimization
        # applied.
        #
        # <
        #    ir_version: 8,
        #    opset_import: ["" : 18]
        # >
        # func (float[1,512] x) => ( split_0,  split_1,  split_2,  split_3) {
        #    int64_128 = Constant <value: tensor = int64 int64_128 {128}> ()
        #    splits = SplitToSequence <axis: int = 1> (x, int64_128)
        #    int64_0 = Constant <value: tensor = int64 int64_0 {0}> ()
        #    split_0 = SequenceAt (splits, int64_0)
        #    int64_1 = Constant <value: tensor = int64 int64_1 {1}> ()
        #    split_1 = SequenceAt (splits, int64_1)
        #    int64_2 = Constant <value: tensor = int64 int64_2 {2}> ()
        #    split_2 = SequenceAt (splits, int64_2)
        #    int64_3 = Constant <value: tensor = int64 int64_3 {3}> ()
        #    split_3 = SequenceAt (splits, int64_3)
        # }
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 2)
        self.assertEqual(len(optimized.graph[-2].outputs), 4)
        self.assertEqual(optimized.graph[-2].op_type, "Split")

    def test_static_split_to_sequence_with_list_split_and_squence_at_is_folded_as_split(
        self,
    ):
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[1,512] x) => (float[1,N] return_val) {
   const = Constant <value: tensor = int64[3] const {256,128,128}> ()
   splits = SplitToSequence <axis: int = 1> (x, const)
   int64_0 = Constant <value: tensor = int64 int64_0 {0}> ()
   split_0 = SequenceAt (splits, int64_0)
   int64_1 = Constant <value: tensor = int64 int64_1 {1}> ()
   split_1 = SequenceAt (splits, int64_1)
   int64_2 = Constant <value: tensor = int64 int64_2 {2}> ()
   split_2 = SequenceAt (splits, int64_2)
   return_val = Concat <axis: int = 1> (split_0, split_1, split_2)
}"""

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 3)
        self.assertEqual(len(optimized.graph[-2].outputs), 3)
        self.assertEqual(optimized.graph[-2].op_type, "Split")

    def test_static_split_to_sequence_with_list_split_no_keepdims_and_squence_at_is_folded_as_split_with_squeeze(
        self,
    ):
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[1,3] x) => (float[1,3] return_val) {
   const = Constant <value: tensor = int64[3] const {1,1,1}> ()
   splits = SplitToSequence <axis: int = 1, keepdims: int = 0> (x, const)
   int64_0 = Constant <value: tensor = int64 int64_0 {0}> ()
   split_0 = SequenceAt (splits, int64_0)
   int64_1 = Constant <value: tensor = int64 int64_1 {1}> ()
   split_1 = SequenceAt (splits, int64_1)
   int64_2 = Constant <value: tensor = int64 int64_2 {2}> ()
   split_2 = SequenceAt (splits, int64_2)
   return_val = Concat <axis: int = 1> (split_0, split_1, split_2)
}"""
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 7)
        self.assertEqual(len(optimized.graph[1].outputs), 3)
        self.assertEqual(optimized.graph[1].op_type, "Split")
        self.assertEqual(len([n for n in optimized.graph if n.op_type == "Squeeze"]), 3)

    def test_split_to_sequence_and_concat_from_sequence_with_new_axis_0(
        self,
    ):
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[1,3] x) => (float[1,3] return_val) {
   const = Constant <value: tensor = int64[3] const {1,1,1}> ()
   splits = SplitToSequence <axis: int = 1> (x, const)
   return_val = ConcatFromSequence <axis: int = 1, new_axis: int = 0> (splits)
}"""

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 3)
        self.assertEqual(optimized.graph[2].op_type, "Concat")

    def test_split_to_sequence_and_concat_from_sequence_with_new_axis_1(
        self,
    ):
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[1,3] x) => (float[1,3] return_val) {
   const = Constant <value: tensor = int64[3] const {1,1,1}> ()
   splits = SplitToSequence <axis: int = 1> (x, const)
   return_val = ConcatFromSequence <axis: int = 1, new_axis: int = 1> (splits)
}"""

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 7)
        self.assertEqual(optimized.graph[6].op_type, "Concat")

    def test_dynamic_split_to_sequence_list_shape_rewrite(self):
        # split is a graph input with known 1-D static shape [4]; values unknown (not constant)
        # Ensures the branch: if isinstance(split_shape, tuple) and len(split_shape) == 1
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[2,N] x, int64[4] split) => (float[2,N] return_val) {
   splits = SplitToSequence <axis: int = 1> (x, split)
   i0 = Constant <value: tensor = int64 i0 {0}> ()
   s0 = SequenceAt (splits, i0)
   i1 = Constant <value: tensor = int64 i1 {1}> ()
   s1 = SequenceAt (splits, i1)
   i2 = Constant <value: tensor = int64 i2 {2}> ()
   s2 = SequenceAt (splits, i2)
   i3 = Constant <value: tensor = int64 i3 {3}> ()
   s3 = SequenceAt (splits, i3)
   return_val = Concat <axis: int = 1> (s0, s1, s2, s3)
}"""
        optimized = self._fold(model)
        # Expect: Split + Concat (index constants & SequenceAt removed)
        split_nodes = [n for n in optimized.graph if n.op_type == "Split"]
        self.assertEqual(len(split_nodes), 1)
        self.assertEqual(len(split_nodes[0].outputs), 4)
        self.assertEqual(split_nodes[0].op_type, "Split")
        self.assertTrue(all(n.op_type != "SequenceAt" for n in optimized.graph))

    def test_dynamic_split_to_sequence_list_shape_no_keepdims(self):
        # keepdims=0 path with dynamic (non-constant) splits input; triggers squeeze logic.
        model = """
<
   ir_version: 8,
   opset_import: ["" : 18]
>
func (float[1,M] x, int64[3] split) => (float[1,M] return_val) {
   splits = SplitToSequence <axis: int = 1, keepdims: int = 0> (x, split)
   i0 = Constant <value: tensor = int64 i0 {0}> ()
   s0 = SequenceAt (splits, i0)
   i1 = Constant <value: tensor = int64 i1 {1}> ()
   s1 = SequenceAt (splits, i1)
   i2 = Constant <value: tensor = int64 i2 {2}> ()
   s2 = SequenceAt (splits, i2)
   return_val = Concat <axis: int = 1> (s0, s1, s2)
}"""
        optimized = self._fold(model)
        split_nodes = [n for n in optimized.graph if n.op_type == "Split"]
        self.assertEqual(len(split_nodes), 1)
        self.assertEqual(len(split_nodes[0].outputs), 3)
        self.assertTrue(all(n.op_type != "SequenceAt" for n in optimized.graph))
        # Each split output should have a corresponding Squeeze (keepdims=0 branch)
        squeeze_nodes = [n for n in optimized.graph if n.op_type == "Squeeze"]
        self.assertEqual(len(squeeze_nodes), 3)

    def test_initializer_input_not_folded(self):
        model_text = """
            <ir_version: 7, opset_import: [ "" : 18]>
            agraph (float[N] x, float[1] c = {1.0} ) => (float[N] z)
            {
                # c is not a constant, and following should not be folded.
                two_c = Add (c, c)
                z = Mul (x, two_c)
            }"""
        optimized = self._fold(model_text)
        self.assertEqual(len(optimized.graph), 2)
        self.assertEqual(optimized.graph.node(0).op_type, "Add")

    @parameterized.parameterized.expand(
        [
            ("output = Dropout(input)",),
            ("output = Dropout(input, zero, true)",),
            ("output = Dropout(input, half)",),
            ("output = Dropout(input, half, false)",),
        ]
    )
    def test_dropout_identity(self, dropout_node: str):
        model = f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] input) => (float[N] output)
            <float zero = {{0.0}}, float half = {{0.5}}, bool true = {{1}}, bool false = {{0}}>
            {{
                {dropout_node}
            }}
        """
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual(optimized.graph.node(0).op_type, "Identity")

    @parameterized.parameterized.expand(
        [
            ("output, mask = Dropout(input)",),
            ("output, mask = Dropout(input, zero, true)",),
            ("output, mask = Dropout(input, half)",),
            ("output, mask = Dropout(input, half, false)",),
        ]
    )
    def test_dropout_identity_mask(self, dropout_node: str):
        model = f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] input) => (float[N] output, bool[N] mask)
            <float zero = {{0.0}}, float half = {{0.5}}, bool true = {{1}}, bool false = {{0}}>
            {{
                {dropout_node}
            }}
        """
        optimized = self._fold(model)
        nodes = list(optimized.graph)
        self.assertEqual(len(nodes), 3)
        ops = [node.op_type for node in nodes]
        self.assertEqual(ops, ["Identity", "Shape", "ConstantOfShape"])

    def test_concat_identity(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z)
            {
                z = Concat <axis=-1> (x)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual(optimized.graph.node(0).op_type, "Identity")

    def test_concat_zero_length(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N, 128] x1, float[N, 0] x2, float[N, 128] x3) => (float[N, M] z)
            {
                z = Concat <axis=-1> (x1, x2, x3)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual([x.name for x in optimized.graph.node(0).inputs], ["x1", "x3"])

    def test_concat_zero_length_identity(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N, 0] x1, float[N, 128] x2, float[N, 0] x3) => (float[N, M] z)
            {
                z = Concat <axis=-1> (x1, x2, x3)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual(optimized.graph.node(0).op_type, "Identity")
        self.assertEqual([x.name for x in optimized.graph.node(0).inputs], ["x2"])

    def test_concat_zero_length_output(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N, 0] x1, float[N, 0] x2, float[N, 0] x3) => (float[N, M] z)
            {
                z = Concat <axis=-1> (x1, x2, x3)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 1)
        self.assertEqual(optimized.graph.node(0).op_type, "Identity")
        self.assertEqual([x.name for x in optimized.graph.node(0).inputs], ["x1"])

    def test_expand_identity(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[128, 256] x) => (float[128, 256] z)
            {
                shape = Constant <value_ints=[128, 256]> ()
                z = Expand (x, shape)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(optimized.graph.node(-1).op_type, "Identity")

    def test_expand_identity_symdim(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[B, 256] x) => (float[B, 256] z)
            {
                b = Shape <start=0, end=1> (x)
                const_256 = Constant <value_ints=[256]> ()
                shape = Concat <axis=0> (b, const_256)
                z = Expand (x, shape)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(optimized.graph.node(-1).op_type, "Identity")

    def test_abs_symdim(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[B, 256] x) => (float[B, 256] z)
            {
                b = Shape <start=0, end=1> (x)
                const_256 = Constant <value_ints=[256]> ()
                b_256 = Concat <axis=0> (b, const_256)
                shape = Abs (b_256)
                z = Expand (x, shape)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(optimized.graph.node(-1).op_type, "Identity")

    def test_reshape_identity(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[128, 256] x) => (float[128, 256] z)
            {
                shape = Constant <value_ints=[128, 256]> ()
                z = Reshape (x, shape)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(optimized.graph.node(-1).op_type, "Identity")

    def test_reshape_identity_symdim(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[B, 256] x, float[B, 128] y) => (float[B, 256] z)
            {
                b = Shape <start=0, end=1> (y)
                const_256 = Constant <value_ints=[256]> ()
                shape = Concat <axis=0> (b, const_256)
                z = Reshape (x, shape)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(optimized.graph.node(-1).op_type, "Identity")

    def test_gather_symdim(self):
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[B, 256] x, float[B, 128] y) => (float[B, 256] z)
            {
                b_128 = Shape (y)
                index_0 = Constant <value_ints=[0]> ()
                b = Gather <axis=0> (b_128, index_0)
                const_256 = Constant <value_ints=[256]> ()
                shape = Concat <axis=0> (b, const_256)
                z = Reshape (x, shape)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(optimized.graph.node(-1).op_type, "Identity")

    def test_input_size_limit(self):
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[M, 256] x) => (float[M, 256] z)
            <float[1, 1] w = {1.0}> # placeholder for large initializer of shape [256, 256]
            {
                w_squared = Mul (w, w)
                z = Add (x, w_squared)
            }
        """
        model = ir.from_onnx_text(model_text)
        w = model.graph.initializers["w"]
        w.shape = ir.Shape([256, 256])
        w.const_value = ir.tensor(np.random.random((256, 256)).astype(np.float32))

        # Input size limit will prevent folding of Mul op
        optimized = self._fold(model, onnx_shape_inference=False, input_size_limit=128 * 128)
        ops = [node.op_type for node in optimized.graph]
        self.assertEqual(ops, ["Mul", "Add"])

        # Input size limit will allow folding of Mul op
        # Since there is no increase in model-size, output-size is not a concern.
        optimized = self._fold(model, input_size_limit=256 * 256, output_size_limit=256 * 256)
        ops = [node.op_type for node in optimized.graph]
        self.assertEqual(ops, ["Add"])
        self.assertIn("w_squared", optimized.graph.initializers)

    def test_transpose_is_always_folded(self):
        model_text = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[M, 256] x) => (float[M, 512] z)
            <float[1, 1] w = {1.0}> # placeholder for large initializer of shape [512, 256]
            {
                z = Transpose (w)
            }
        """
        model = ir.from_onnx_text(model_text)
        w = model.graph.initializers["w"]
        w.shape = ir.Shape([512, 256])
        w.const_value = ir.tensor(np.random.random((512, 256)).astype(np.float32))

        # Input size limit will not prevent folding of Transpose op
        optimized = self._fold(model, input_size_limit=1)
        ops = [node.op_type for node in optimized.graph]
        self.assertEqual(ops, [])
        self.assertIn("z", optimized.graph.initializers)

    def test_node_is_folded_if_specified_as_should_fold(self):
        model_text = """
            <ir_version: 10, opset_import: [ "" : 20]>
            agraph (float[M, 256] x) => (float[42, 42] z)
            <int64[2] w = {42, 42}>
            {
                z = ConstantOfShape <value: tensor = int64[1] {1}> (w)
            }
        """
        model = ir.from_onnx_text(model_text)

        # ConstantOfShape is not folded by default
        optimized = self._fold(model)
        ops = [node.op_type for node in optimized.graph]
        self.assertEqual(ops, ["ConstantOfShape"])

        # But ConstantOfShape is folded when specified in should_fold
        optimized = self._fold(
            model, should_fold=lambda node: node.op_type == "ConstantOfShape" or None
        )
        ops = [node.op_type for node in optimized.graph]
        self.assertEqual(ops, [])
        self.assertIn("z", optimized.graph.initializers)
        np.testing.assert_array_equal(
            optimized.graph.initializers["z"].const_value,
            np.ones((42, 42), dtype=np.int64),
        )

    def test_quantize_linear_is_not_folded(self):
        model_text = """
            <ir_version: 10, opset_import: [ "" : 20]>
            agraph () => (uint8[4] z)
            <float[4] x = {1.0, 2.0, 3.0, 4.0}, float[1] scale = {1.0}, uint8[1] zero_point = {0}>
            {
                z = QuantizeLinear (x, scale, zero_point)
            }
        """
        model = ir.from_onnx_text(model_text)
        optimized = self._fold(model)
        ops = [node.op_type for node in optimized.graph]
        # QuantizeLinear should not be folded even when all inputs are constants
        self.assertEqual(ops, ["QuantizeLinear"])

    def test_dequantize_linear_is_not_folded(self):
        model_text = """
            <ir_version: 10, opset_import: [ "" : 20]>
            agraph () => (float[4] z)
            <uint8[4] x = {1, 2, 3, 4}, float[1] scale = {1.0}, uint8[1] zero_point = {0}>
            {
                z = DequantizeLinear (x, scale, zero_point)
            }
        """
        model = ir.from_onnx_text(model_text)
        optimized = self._fold(model)
        ops = [node.op_type for node in optimized.graph]
        # DequantizeLinear should not be folded even when all inputs are constants
        self.assertEqual(ops, ["DequantizeLinear"])

    def test_multi_graph_identity_output_preserves_output_name(self):
        model = """
            <ir_version: 10, opset_import: ["" : 20]>
            agraph (float[N] x) => (float[N] graph_output1, float[N] graph_output2) {
                t = Identity(x)
                graph_output1 = Identity(t)
                graph_output2 = Identity(t)
            }"""
        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 2)
        self.assertEqual([n.op_type for n in optimized.graph], ["Identity", "Identity"])
        self.assertEqual(
            [n.outputs[0].name for n in optimized.graph], ["graph_output1", "graph_output2"]
        )
        self.assertEqual([input.name for input in optimized.graph.inputs], ["x"])

    # This should not be constant-foldable as the constant references an
    # attribute and thus the shape cannot be resolved. At the same time it
    # should not fail due to the attribute value being None in
    # _process_constant_node
    def test_attribute_reference(self):
        model = """
            <ir_version: 7, opset_import: ["" : 17]>
            agraph () => (int64[N] z) {
                x = Constant <value_ints: ints = @attr> ()
                z = Shape (x)
            }
        """

        optimized = self._fold(model)
        self.assertEqual(len(optimized.graph), 2)

    def test_constant_folding_creates_constant_nodes_in_function(self):
        model = """
            <ir_version: 9, opset_import: ["this" : 1, "" : 19]>
            model (float x) => (float return_val) {
                return_val = this.function (x)
            }
            <domain: "this", opset_import: ["" : 19]>
            function (x) => (return_val) {
                tmp = Constant <value_int=1> ()
                tmp_0 = Cast <to=1> (tmp)
                return_val = Sub (tmp_0, x)
            }
        """
        optimized = self._fold(model)
        self.assertEqual(len(optimized.functions), 1)
        for func in optimized.functions.values():
            # Ensure that constant folding has created constant nodes in the function
            constant_nodes = [n for n in func.graph if n.op_type == "Constant"]
            self.assertEqual(len(constant_nodes), 1)

    def test_initializer_as_graph_output_is_not_removed(self):
        """Test that an initializer that is a graph output is not removed during constant folding."""
        model = """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] y, float z) {
                constant = Constant <value_float=2.0> ()
                y = Mul(x, constant)
                z = Identity(constant)
            }
        """

        optimized = self._fold(model)
        # After constant folding, the Identity node should be folded, and 'constant'
        # should become an initializer with the output name 'z'.
        # The key thing is that this initializer should NOT be removed even though
        # the Identity node was folded, because it is a graph output.
        self.assertIn("z", optimized.graph.initializers)
        # The Identity node should be removed
        identity_nodes = [n for n in optimized.graph if n.op_type == "Identity"]
        self.assertEqual(len(identity_nodes), 0)
        # Verify the graph still has both outputs
        output_names = [o.name for o in optimized.graph.outputs]
        self.assertIn("y", output_names)
        self.assertIn("z", output_names)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import os
import sys
import time
import unittest

import numpy as np
import torch

import onnxscript.tools.memory_peak


class TestMemoryPeak(unittest.TestCase):
    @unittest.skipIf(sys.platform == "win32", reason="other test are failing")
    def test_memory(self):
        mem = onnxscript.tools.memory_peak.get_memory_rss(os.getpid())
        self.assertIsInstance(mem, int)

    @unittest.skipIf(sys.platform == "win32", reason="other test are failing")
    def test_spy(self):
        p = onnxscript.tools.memory_peak.start_spying_on()
        res = []
        for i in range(10):
            time.sleep(0.005)
            res.append(np.empty(i * 1000000))
        del res
        time.sleep(0.02)
        pres = p.stop()
        self.assertIsInstance(pres, dict)
        self.assertLessEqual(pres["cpu"][0].end, pres["cpu"][0].max_peak)
        self.assertLessEqual(pres["cpu"][0].begin, pres["cpu"][0].max_peak)
        self.assertIsInstance(pres["cpu"][0].to_dict(), dict)

    @unittest.skipIf(not torch.cuda.is_available(), reason="CUDA not here")
    def test_spy_cuda(self):
        p = onnxscript.tools.memory_peak.start_spying_on(cuda=True)
        res = []
        for i in range(10):
            time.sleep(0.005)
            res.append(np.empty(i * 1000000))
        del res
        time.sleep(0.02)
        pres = p.stop()
        self.assertIsInstance(pres, dict)
        self.assertIsInstance(pres["cpu"], list)
        self.assertEqual(len(pres["cpu"]), 1)
        self.assertIsInstance(pres["gpus"], list)
        self.assertLessEqual(pres["cpu"][0].end, pres["cpu"][0].max_peak)
        self.assertLessEqual(pres["cpu"][0].begin, pres["cpu"][0].max_peak)
        self.assertIn("gpus", pres)
        self.assertLessEqual(pres["gpus"][0].end, pres["gpus"][0].max_peak)
        self.assertLessEqual(pres["gpus"][0].begin, pres["gpus"][0].max_peak)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D214, D402, D405, D411, D416
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, TypeVar, Union

from onnx import GraphProto
from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset15 import Opset15
from onnxscript.onnx_types import (
    BFLOAT16,
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset16(Opset15):
    def __new__(cls):
        return Opset.__new__(cls, "", 16)

    T_GreaterOrEqual = TypeVar(
        "T_GreaterOrEqual",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_GreaterOrEqual: TypeAlias = BOOL

    def GreaterOrEqual(self, A: T_GreaterOrEqual, B: T_GreaterOrEqual) -> T1_GreaterOrEqual:
        r"""[🌐 GreaterOrEqual(16)](https://onnx.ai/onnx/operators/onnx__GreaterOrEqual.html#greaterorequal-16 "Online Documentation")


        Returns the tensor resulted from performing the `greater_equal` logical operation
        elementwise on the input tensors `A` and `B` (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            A: (non-differentiable) First input operand for the logical operator.

            B: (non-differentiable) Second input operand for the logical operator.
        """

        schema = get_schema("GreaterOrEqual", 16, "")
        op = Op(self, "GreaterOrEqual", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T1_GridSample = TypeVar(
        "T1_GridSample",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T2_GridSample = TypeVar("T2_GridSample", DOUBLE, FLOAT, FLOAT16)

    def GridSample(
        self,
        X: T1_GridSample,
        grid: T2_GridSample,
        *,
        align_corners: int = 0,
        mode: str = "bilinear",
        padding_mode: str = "zeros",
    ) -> T1_GridSample:
        r"""[🌐 GridSample(16)](https://onnx.ai/onnx/operators/onnx__GridSample.html#gridsample-16 "Online Documentation")


        Given an input `X` and a flow-field `grid`, computes the output `Y` using `X` values and pixel locations from `grid`.
        Currently, only spatial (4-D) inputs are supported. For input `X` with shape (N, C, H, W) and `grid` with shape (N, H_out, W_out, 2),
        the output `Y` will have shape (N, C, H_out, W_out).

        The tensor `X` contains values at centers of square pixels in a H by W 2-dimensional image.
        The tensor `grid` describes normalized positions where the output `Y` is to be computed
        using a specified interpolation method (the mode) and a padding mode (for grid positions falling outside the 2-dimensional image).

        Elements in `grid[N, H_out, W_out]` are size-2 vectors specifying positions in the 2-dimensional space of `X`.
        They are used to interpolate output values of `Y[N, C, H_out, W_out]`.

        The GridSample operator is often used in doing grid generator and sampler in the [Spatial Transformer Networks](https://arxiv.org/abs/1506.02025).
        See also in [torch.nn.functional.grid_sample](https://pytorch.org/docs/master/generated/torch.nn.functional.grid_sample.html#torch-nn-functional-grid-sample).


        Args:
            X: (differentiable) 4-D tensor of shape (N, C, H, W), where N is the batch
                size, C is the numbers of channels, H and W are the height and width of
                the input data.

            grid: (non-differentiable) Input offset, 4-D tensor of shape (N, H_out,
                W_out, 2), where H_out and W_out are the height and width of grid and
                output, Grid specifies the sampling pixel locations normalized by the
                input spatial dimensions. Therefore, it should have most values in the
                range of [-1, 1]. If grid has values outside the range of [-1, 1], the
                corresponding outputs will be handled as defined by padding_mode.

            align_corners: If align_corners=1, the extrema (-1 and 1) are considered as
                referring to the center points of the input's corner pixels. If
                align_corners=0, they are instead considered as referring to the corner
                points of the input's corner pixels, making the sampling more resolution
                agnostic.

            mode: Three interpolation modes: bilinear (default), nearest and bicubic.

            padding_mode: Support padding modes for outside grid values:
                `zeros`(default), `border`, `reflection`. zeros: use 0 for out-of-bound
                grid locations, border: use border values for out-of-bound grid
                locations, reflection: use values at locations reflected by the border
                for out-of-bound grid locations. If index 0 represents the margin pixel,
                the reflected value at index -1 will be the same as the value at index
                1. For location far away from the border, it will keep being reflected
                until becoming in bound. If pixel location x = -3.5 reflects by border
                -1 and becomes x' = 1.5, then reflects by border 1 and becomes x'' =
                0.5.
        """

        schema = get_schema("GridSample", 16, "")
        op = Op(self, "GridSample", schema)
        return op(
            *self._prepare_inputs(schema, X, grid),
            align_corners=align_corners,
            mode=mode,
            padding_mode=padding_mode,
        )

    V_Identity = TypeVar(
        "V_Identity",
        Optional[Sequence[BOOL]],
        Optional[Sequence[COMPLEX128]],
        Optional[Sequence[COMPLEX64]],
        Optional[Sequence[DOUBLE]],
        Optional[Sequence[FLOAT]],
        Optional[Sequence[FLOAT16]],
        Optional[Sequence[INT16]],
        Optional[Sequence[INT32]],
        Optional[Sequence[INT64]],
        Optional[Sequence[INT8]],
        Optional[Sequence[STRING]],
        Optional[Sequence[UINT16]],
        Optional[Sequence[UINT32]],
        Optional[Sequence[UINT64]],
        Optional[Sequence[UINT8]],
        Optional[BOOL],
        Optional[COMPLEX128],
        Optional[COMPLEX64],
        Optional[DOUBLE],
        Optional[FLOAT],
        Optional[FLOAT16],
        Optional[INT16],
        Optional[INT32],
        Optional[INT64],
        Optional[INT8],
        Optional[STRING],
        Optional[UINT16],
        Optional[UINT32],
        Optional[UINT64],
        Optional[UINT8],
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Identity(self, input: V_Identity) -> V_Identity:
        r"""[🌐 Identity(16)](https://onnx.ai/onnx/operators/onnx__Identity.html#identity-16 "Online Documentation")

        Identity operator

        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Identity", 16, "")
        op = Op(self, "Identity", schema)
        return op(*self._prepare_inputs(schema, input))

    B_If: TypeAlias = BOOL

    V_If: TypeAlias = Union[
        None,
        Sequence[BFLOAT16],
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    ]

    def If(self, cond: B_If, *, else_branch: GraphProto, then_branch: GraphProto) -> V_If:
        r"""[🌐 If(16)](https://onnx.ai/onnx/operators/onnx__If.html#if-16 "Online Documentation")

        If conditional

        Args:
            cond: Condition for the if. The tensor must contain a single element.

            else_branch: Graph to run if condition is false. Has N outputs: values you
                wish to be live-out to the enclosing scope. The number of outputs must
                match the number of outputs in the then_branch.

            then_branch: Graph to run if condition is true. Has N outputs: values you
                wish to be live-out to the enclosing scope. The number of outputs must
                match the number of outputs in the else_branch.
        """

        schema = get_schema("If", 16, "")
        op = Op(self, "If", schema)
        return op(
            *self._prepare_inputs(schema, cond),
            else_branch=else_branch,
            then_branch=then_branch,
        )

    T_LeakyRelu = TypeVar("T_LeakyRelu", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    def LeakyRelu(self, X: T_LeakyRelu, *, alpha: float = 0.009999999776482582) -> T_LeakyRelu:
        r"""[🌐 LeakyRelu(16)](https://onnx.ai/onnx/operators/onnx__LeakyRelu.html#leakyrelu-16 "Online Documentation")


        LeakyRelu takes input data (Tensor<T>) and an argument alpha, and produces one
        output data (Tensor<T>) where the function `f(x) = alpha * x for x < 0`,
        `f(x) = x for x >= 0`, is applied to the data tensor elementwise.


        Args:
            X: (differentiable) Input tensor

            alpha: Coefficient of leakage.
        """

        schema = get_schema("LeakyRelu", 16, "")
        op = Op(self, "LeakyRelu", schema)
        return op(*self._prepare_inputs(schema, X), alpha=alpha)

    T_LessOrEqual = TypeVar(
        "T_LessOrEqual",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    T1_LessOrEqual: TypeAlias = BOOL

    def LessOrEqual(self, A: T_LessOrEqual, B: T_LessOrEqual) -> T1_LessOrEqual:
        r"""[🌐 LessOrEqual(16)](https://onnx.ai/onnx/operators/onnx__LessOrEqual.html#lessorequal-16 "Online Documentation")


        Returns the tensor resulted from performing the `less_equal` logical operation
        elementwise on the input tensors `A` and `B` (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            A: (non-differentiable) First input operand for the logical operator.

            B: (non-differentiable) Second input operand for the logical operator.
        """

        schema = get_schema("LessOrEqual", 16, "")
        op = Op(self, "LessOrEqual", schema)
        return op(*self._prepare_inputs(schema, A, B))

    I_Loop: TypeAlias = INT64

    B_Loop: TypeAlias = BOOL

    V_Loop = TypeVar(
        "V_Loop",
        Optional[Sequence[BFLOAT16]],
        Optional[Sequence[BOOL]],
        Optional[Sequence[COMPLEX128]],
        Optional[Sequence[COMPLEX64]],
        Optional[Sequence[DOUBLE]],
        Optional[Sequence[FLOAT]],
        Optional[Sequence[FLOAT16]],
        Optional[Sequence[INT16]],
        Optional[Sequence[INT32]],
        Optional[Sequence[INT64]],
        Optional[Sequence[INT8]],
        Optional[Sequence[STRING]],
        Optional[Sequence[UINT16]],
        Optional[Sequence[UINT32]],
        Optional[Sequence[UINT64]],
        Optional[Sequence[UINT8]],
        Optional[BFLOAT16],
        Optional[BOOL],
        Optional[COMPLEX128],
        Optional[COMPLEX64],
        Optional[DOUBLE],
        Optional[FLOAT],
        Optional[FLOAT16],
        Optional[INT16],
        Optional[INT32],
        Optional[INT64],
        Optional[INT8],
        Optional[STRING],
        Optional[UINT16],
        Optional[UINT32],
        Optional[UINT64],
        Optional[UINT8],
        Sequence[BFLOAT16],
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Loop(
        self,
        M: Optional[I_Loop],
        cond: Optional[B_Loop],
        *v_initial: V_Loop,
        body: GraphProto,
    ) -> V_Loop:
        r"""[🌐 Loop(16)](https://onnx.ai/onnx/operators/onnx__Loop.html#loop-16 "Online Documentation")


        Generic Looping construct. This loop has multiple termination conditions:

        1) Trip count. Iteration count specified at runtime. Set by
           specifying the input M. Optional. Set to empty string to omit.
           Note that a static trip count (specified at graph construction time) can be
           specified by passing in a constant node for input M.
        2) Loop termination condition. This is an input to the op that determines
           whether to run the first iteration and also a loop-carried dependency for
           the body graph. The body graph must yield a value for the condition variable,
           whether this input is provided or not.

        This table summarizes the operating modes of this operator with equivalent
        C-style code:

        Operator inputs defined as (max_trip_count, condition_var).

        * input ("", ""):
                for (int i=0; ; ++i) {
                  cond = ... // Note this value is ignored, but is required in the body
                }

        * input ("", cond) // Note this is analogous to a while loop
                bool cond = ...;
                for (int i=0; cond; ++i) {
                  cond = ...;
                }

        * input ("", 1) // Note this is analogous to a do-while loop
                bool cond = true
                for (int i=0; cond; ++i) {
                  cond = ...;
                }

        * input (trip_count, "") // Note this is analogous to a for loop
                int trip_count = ...
                for (int i=0; i < trip_count; ++i) {
                  cond = ...; // ignored
                }

        * input (trip_count, cond)
                int trip_count = ...;
                bool cond = ...;
                for (int i=0; i < trip_count && cond; ++i) {
                  cond = ...;
                }


        *Sample usage - cond as well as trip count*

            graph predict-net {
              %a = Constant[value = <Scalar Tensor [3]>]()
              %b = Constant[value = <Scalar Tensor [6]>]()
              %keepgoing = Constant[value = <Scalar Tensor [1]>]()
              %max_trip_count = Constant[value = <Scalar Tensor [10]>]()
              %keepgoing_out, %b_out, %user_defined_vals = Loop[body = <graph body-net>](%max_trip_count, %keepgoing, %b)
              return
            }

            graph body-net (
              %i[INT32, scalar]           // iteration number
              %keepgoing_in[BOOL, scalar] // incoming loop-termination-condition; not used
              %b_in[INT32, scalar]        // incoming value of loop-carried-dependency b
            ) {
              %my_local = Add(%a, %b_in)
              %b_out = Sub(%a, %b_in) // outgoing value of loop-carried-dependency b
              %keepgoing_out = Greater(%my_local, %b_out) // outgoing loop-termination-condition
              %user_defined_val = Add(%b_in, %b_in) // scan-output value to be accumulated
              return %keepgoing_out, %b_out, %user_defined_val
            }

        *Sample equivalent C code*

            {
              /* User-defined code (enclosing scope) */
              int a = 3, b = 6;
              bool keepgoing = true; // Analogous to input cond
              /* End user-defined code */

              /* Implicitly-defined code */
              const int max_trip_count = 10; // Analogous to input M
              int user_defined_vals[]; // Imagine this is resizable
              /* End implicitly-defined code */
              /* initialize loop-carried variables and scan-output variables */
              bool keepgoing_out = keepgoing
              int b_out = b

              for (int i=0; i < max_trip_count && keepgoing_out; ++i) {
                /* Implicitly-defined code: bind actual parameter values
                   to formal parameter variables of loop-body */
                bool keepgoing_in = keepgoing_out;
                bool b_in = b_out;

                /* User-defined code (loop body) */
                int my_local = a + b_in; // Reading value "a" from the enclosing scope is fine
                b_out = a - b_in;
                keepgoing_out = my_local > b_out;
                user_defined_val = b_in + b_in; // b_in and b_out are different variables
                /* End user-defined code */

                /* Implicitly defined-code */
                user_defined_vals[i] = user_defined_val // accumulate scan-output values
              }
              // int t = my_local; // Can't do this. my_local is not accessible here.

              // The values below are bound to the output variables of the loop and therefore accessible
              // b_out; user_defined_vals; keepgoing_out;
            }

        There are several things of note in this code snippet:

        1) Values from the enclosing scope (i.e. variable "a" here) are in scope and can
           be referenced in the inputs of the loop.
        2) Any values computed in the loop body that needs to be used in a subsequent
           iteration or after the loop are modelled using a pair of variables in the loop-body,
           consisting of an input variable (eg., b_in) and an output variable (eg., b_out).
           These are referred to as loop-carried dependences. The loop operation node
           supplies the input value of the input variable for the first iteration, and
           returns the output value of the output variable produced by the final
           iteration.
        3) Scan_output variables are used to implicitly concatenate values computed across
           all the iterations. In the above example, the value of user_defined_val computed
           over all iterations are concatenated and returned as the value of user_defined_vals
           after the loop.
        4) Values created in the body cannot be accessed in the enclosing scope,
           except using the mechanism described above.

        Note that the semantics of this op support "diagonal" or "wavefront" execution.
        (See Step 3 here for an example:
        https://devblogs.nvidia.com/optimizing-recurrent-neural-networks-cudnn-5/).
        Frontends should emit multi-layer RNNs as a series of While operators (with
        time being the inner looping dimension), with each successive layer consuming
        the scan_outputs from the previous layer, possibly going through several
        point-wise operators (e.g. dropout, residual connections, linear layer).

        The input/output of subgraph (produced by loop node) matching is based on order instead of name. The implementation will figure out the names based on this order.


        Args:
            M: (optional) A maximum trip-count for the loop specified at runtime.
                Optional. Pass empty string to skip.

            cond: (optional) A boolean termination condition. Optional. Pass empty
                string to skip.

            v_initial: (variadic, heterogeneous) The initial values of any loop-carried
                dependencies (values that change across loop iterations)

            body: The graph run each iteration. It has 2+N inputs: (iteration_num,
                condition, loop carried dependencies...). It has 1+N+K outputs:
                (condition, loop carried dependencies..., scan_outputs...). Each
                scan_output is created by concatenating the value of the specified
                output value at the end of each iteration of the loop. It is an error if
                the dimensions or data type of these scan_outputs change across loop
                iterations.
        """

        schema = get_schema("Loop", 16, "")
        op = Op(self, "Loop", schema)
        return op(*self._prepare_inputs(schema, M, cond, *v_initial), body=body)

    T_PRelu = TypeVar(
        "T_PRelu", BFLOAT16, DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64
    )

    def PRelu(self, X: T_PRelu, slope: T_PRelu) -> T_PRelu:
        r"""[🌐 PRelu(16)](https://onnx.ai/onnx/operators/onnx__PRelu.html#prelu-16 "Online Documentation")


        PRelu takes input data (Tensor<T>) and slope tensor as input, and produces one
        output data (Tensor<T>) where the function `f(x) = slope * x for x < 0`,
        `f(x) = x for x >= 0`., is applied to the data tensor elementwise.
        This operator supports **unidirectional broadcasting** (tensor slope should be unidirectional broadcastable to input tensor X); for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        Args:
            X: (differentiable) Input tensor

            slope: (differentiable) Slope tensor. The shape of slope can be smaller than
                first input X; if so, its shape must be unidirectional broadcastable to
                X
        """

        schema = get_schema("PRelu", 16, "")
        op = Op(self, "PRelu", schema)
        return op(*self._prepare_inputs(schema, X, slope))

    T1_RoiAlign = TypeVar("T1_RoiAlign", DOUBLE, FLOAT, FLOAT16)

    T2_RoiAlign: TypeAlias = INT64

    def RoiAlign(
        self,
        X: T1_RoiAlign,
        rois: T1_RoiAlign,
        batch_indices: T2_RoiAlign,
        *,
        coordinate_transformation_mode: str = "half_pixel",
        mode: str = "avg",
        output_height: int = 1,
        output_width: int = 1,
        sampling_ratio: int = 0,
        spatial_scale: float = 1.0,
    ) -> T1_RoiAlign:
        r"""[🌐 RoiAlign(16)](https://onnx.ai/onnx/operators/onnx__RoiAlign.html#roialign-16 "Online Documentation")


        Region of Interest (RoI) align operation described in the
        [Mask R-CNN paper](https://arxiv.org/abs/1703.06870).
        RoiAlign consumes an input tensor X and region of interests (rois)
        to apply pooling across each RoI; it produces a 4-D tensor of shape
        (num_rois, C, output_height, output_width).

        RoiAlign is proposed to avoid the misalignment by removing
        quantizations while converting from original image into feature
        map and from feature map into RoI feature; in each ROI bin,
        the value of the sampled locations are computed directly
        through bilinear interpolation.


        Args:
            X: Input data tensor from the previous operator; 4-D feature map of shape
                (N, C, H, W), where N is the batch size, C is the number of channels,
                and H and W are the height and the width of the data.

            rois: RoIs (Regions of Interest) to pool over; rois is 2-D input of shape
                (num_rois, 4) given as [[x1, y1, x2, y2], ...]. The RoIs' coordinates
                are in the coordinate system of the input image. Each coordinate set has
                a 1:1 correspondence with the 'batch_indices' input.

            batch_indices: 1-D tensor of shape (num_rois,) with each element denoting
                the index of the corresponding image in the batch.

            coordinate_transformation_mode: Allowed values are 'half_pixel' and
                'output_half_pixel'. Use the value 'half_pixel' to pixel shift the input
                coordinates by -0.5 (the recommended behavior). Use the value
                'output_half_pixel' to omit the pixel shift for the input (use this for
                a backward-compatible behavior).

            mode: The pooling method. Two modes are supported: 'avg' and 'max'. Default
                is 'avg'.

            output_height: default 1; Pooled output Y's height.

            output_width: default 1; Pooled output Y's width.

            sampling_ratio: Number of sampling points in the interpolation grid used to
                compute the output value of each pooled output bin. If > 0, then exactly
                sampling_ratio x sampling_ratio grid points are used. If == 0, then an
                adaptive number of grid points are used (computed as ceil(roi_width /
                output_width), and likewise for height). Default is 0.

            spatial_scale: Multiplicative spatial scale factor to translate ROI
                coordinates from their input spatial scale to the scale used when
                pooling, i.e., spatial scale of the input feature map X relative to the
                input image. E.g.; default is 1.0f.
        """

        schema = get_schema("RoiAlign", 16, "")
        op = Op(self, "RoiAlign", schema)
        return op(
            *self._prepare_inputs(schema, X, rois, batch_indices),
            coordinate_transformation_mode=coordinate_transformation_mode,
            mode=mode,
            output_height=output_height,
            output_width=output_width,
            sampling_ratio=sampling_ratio,
            spatial_scale=spatial_scale,
        )

    V_Scan = TypeVar(
        "V_Scan",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Scan(
        self,
        *initial_state_and_scan_inputs: V_Scan,
        body: GraphProto,
        num_scan_inputs: int,
        scan_input_axes: Optional[Sequence[int]] = None,
        scan_input_directions: Optional[Sequence[int]] = None,
        scan_output_axes: Optional[Sequence[int]] = None,
        scan_output_directions: Optional[Sequence[int]] = None,
    ) -> V_Scan:
        r"""[🌐 Scan(16)](https://onnx.ai/onnx/operators/onnx__Scan.html#scan-16 "Online Documentation")


        Scan can be used to iterate over one or more scan_input tensors,
        constructing zero or more scan_output tensors. It combines ideas from general recurrences,
        functional programming constructs such as scan, fold, map, and zip, and is intended to enable
        generalizations of RNN-like constructs for sequence-to-sequence processing.
        Other tensors (referred to as state_variables here) can be used to carry a state
        when iterating from one element to another (similar to hidden-state in RNNs, also referred
        to as loop-carried dependences in the context of loops).
        Many common usages involve a single scan_input tensor (where functionality
        similar to scan, fold and map can be obtained). When more than one scan_input is used,
        a behavior similar to zip is obtained.

        The attribute body must be a graph, specifying the computation to be performed in
        every iteration. It takes as input the current values of the state_variables and
        the current iterated element of the scan_inputs. It must return the (updated) values
        of the state_variables and zero or more scan_output_element tensors. The values of the
        scan_output_element tensors are concatenated over all the iterations to produce the
        scan_output values of the scan construct (similar to the concatenated intermediate
        hidden-state values of RNN-like constructs). All the output tensors (state_variables as
        well as scan_output_element tensors) are required to have the same shape in each iteration
        of the loop (a restriction imposed to enable efficient memory allocation).

        Note that the iterated element passed to the body subgraph does not have a sequence
        axis. It will have a rank one less than the rank of the corresponding scan_input.

        The scan operation returns the final values of the state_variables as well as the
        scan_outputs.

        The optional attribute scan_input_directions specifies the direction (forward or backward)
        for each scan input. If this attribute is omitted, all sequences are scanned in the forward
        direction. A bidirectional scan may be performed by specifying the same tensor input twice
        in the scan_inputs, once with a forward direction, and once with a backward direction.

        The scan_output of the operation is produced by concatenating the scan_output_element
        values produced by the body in each iteration.  The optional attribute scan_output_directions
        specifies the direction in which scan_output is constructed (by appending or prepending the
        scan_output_element to scan_output in each iteration) for each scan_output. If this attribute
        is omitted, the scan_output_element is appended to the scan_output in each iteration.

        The optional attribute scan_input_axes specifies the axis to be scanned for each scan_input.
        If omitted, every scan_input will be scanned in axis 0. For example, if axis 0 is the
        batch axis and axis 1 is the time axis (to be scanned), specify an axis value of 1.
        Note that scanning a non-zero axis may be less efficient than scanning axis zero.

        The optional attribute scan_output_axes specifies the axis along which the scan_outputs
        are accumulated for each scan_output. For example, if axis 1 is the time axis (to be
        scanned) for both inputs and outputs, specify a scan_input axis and scan_output axis
        value of 1.

        Note that because of the ONNX restriction that only the last parameter of an operator can
        be variadic, the initial-states and scan-inputs are listed together as one input parameter.
        Similarly, the final-states and scan-outputs are listed together as one output parameter.
        The attribute num_scan_inputs indicates the number M of scan-inputs.

        The behavior of

            Scan <
                num_scan_inputs = m,
                body = loop-body,
                scan_input_axes = [axis_1, ..., axis_m]
            > (init_1, ..., init_n, scan_1, ..., scan_m)

        is equivalent to the following pseudo-code:

            // scan_i.shape[axis_i] denotes the (max) sequence-length of scan_i
            // scan_i.shape[axis_i] is required to be equal to scan_j.shape[axis_j] for all i,j.
            sequence_length = scan_1.shape[axis_1];

            // initialize state-variables
            st_1 = init_1; ... st_n = init_n;
            // initialize scan-output variables: [] denotes an empty tensor
            scan_out_1 = []; ...; scan_out_k = [];
            // identify number of iterations:

            // execute loop
            for (int t = 0; t < sequence_length; ++t) {
                // generate the scan-input elements: the notation T<axis=k>[t] indicates the sub-tensor
                // of rank one less than T obtained by indexing T at position t along axis k.
                si_1 = scan_1<axis=axis_1>[t];
                ... ;
                si_m = scan_m<axis=axis_m>[t];
                // execute loop-body
                st_1, ..., st_n, so_1, ..., so_k = loop-body(st_1, ..., st_n, si_1, ..., si_m)
                // accumulate the scan-output elements
                scan_out_1 = Concat<axis=0>(scan_out_1, so_1); ... ; scan_out_k = Concat<axis=0>(scan_out_k, so_k);
            }

            return st_1, ..., st_n, scan_out_1, ..., scan_out_k;

        *Sample usage: Encoding RNN using a Scan*

        The following example shows how a simple RNN over an input tensor %X, with weight tensor %Wi,
        recurrence weight tensor %Ri, bias tensors %Wbi and %Rbi, and initial hidden-state %H_0 can
        be encoded as a ScanLoop. Note that the loop-body is a nested graph, and it directly computes
        %Wi, %Ri, %Wbi, and %Rbi (typically constants or initializers in the body graph). If these
        values are computed in the outer graph, they need to be passed in as extra state_variables.

            graph rnn-encoding {
              %H_0 = ...
              %X = ...
              %Y_h, %Y = Scan[body = <graph rnn-cell-1>, num_scan_inputs=1](%H_0, %X)
              return %Y, %Y_h
            }

            graph rnn-cell-1 (
              %H_tminus1[FLOAT, tensor]
              %X_t[FLOAT, tensor]
            ) {
              %Wi = ...
              %Ri = ...
              %Wbi = ...
              %Rbi = ...
              %t1 = X_t * (Wi^T)
              %t2 = H_tminus1*(Ri^T)
              %t3 = Add(%t1, %t2)
              %t4 = Add(%t3, %Wbi)
              %t5 = Add(%t4, %Rbi)
              %Ht = Tanh(%t5)
              %Accumulate = Identity(%Ht)
              return %Ht, %Accumulate
            }



        Args:
            initial_state_and_scan_inputs: (variadic, heterogeneous) Initial values of
                the loop's N state variables followed by M scan_inputs

            body: The graph run each iteration. It has N+M inputs: (loop state
                variables..., scan_input_elts...). It has N+K outputs: (loop state
                variables..., scan_output_elts...). Each scan_output is created by
                concatenating the value of the specified scan_output_elt value at the
                end of each iteration of the loop. It is an error if the dimensions of
                these values change across loop iterations.

            num_scan_inputs: An attribute specifying the number of scan_inputs M.

            scan_input_axes: An optional list of M flags. The i-th element of the list
                specifies the axis to be scanned (the sequence axis) for the i-th
                scan_input. If omitted, 0 will be used as the scan axis for every
                scan_input. Negative value for an axis means counting dimensions from
                the back. Accepted range is [-r, r-1] where r = rank(input).

            scan_input_directions: An optional list of M flags. The i-th element of the
                list specifies the direction to be scanned for the i-th scan_input
                tensor: 0 indicates forward direction and 1 indicates reverse direction.
                If omitted, all scan_input tensors will be scanned in the forward
                direction.

            scan_output_axes: An optional list of K flags. The i-th element of the list
                specifies the axis for the i-th scan_output. The scan outputs are
                accumulated along the specified axis. If omitted, 0 will be used as the
                scan axis for every scan_output. Negative value for an axis means
                counting dimensions from the back. Accepted range is [-r, r-1].

            scan_output_directions: An optional list of K flags, one for each
                scan_output. The i-th element of the list specifies whether the i-th
                scan_output should be constructed by appending or prepending a new value
                in each iteration: 0 indicates appending and 1 indicates prepending. If
                omitted, all scan_output tensors will be produced by appending a value
                in each iteration.
        """

        schema = get_schema("Scan", 16, "")
        op = Op(self, "Scan", schema)
        return op(
            *self._prepare_inputs(schema, *initial_state_and_scan_inputs),
            body=body,
            num_scan_inputs=num_scan_inputs,
            scan_input_axes=scan_input_axes,
            scan_input_directions=scan_input_directions,
            scan_output_axes=scan_output_axes,
            scan_output_directions=scan_output_directions,
        )

    T_ScatterElements = TypeVar(
        "T_ScatterElements",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    Tind_ScatterElements = TypeVar("Tind_ScatterElements", INT32, INT64)

    def ScatterElements(
        self,
        data: T_ScatterElements,
        indices: Tind_ScatterElements,
        updates: T_ScatterElements,
        *,
        axis: int = 0,
        reduction: str = "none",
    ) -> T_ScatterElements:
        r"""[🌐 ScatterElements(16)](https://onnx.ai/onnx/operators/onnx__ScatterElements.html#scatterelements-16 "Online Documentation")


        ScatterElements takes three inputs `data`, `updates`, and `indices` of the same
        rank r >= 1 and an optional attribute axis that identifies an axis of `data`
        (by default, the outer-most axis, that is axis 0). The output of the operation
        is produced by creating a copy of the input `data`, and then updating its value
        to values specified by `updates` at specific index positions specified by
        `indices`. Its output shape is the same as the shape of `data`.
        For each entry in `updates`, the target index in `data` is obtained by combining
        the corresponding entry in `indices` with the index of the entry itself: the
        index-value for dimension = axis is obtained from the value of the corresponding
        entry in `indices` and the index-value for dimension != axis is obtained from the
        index of the entry itself.
        `reduction` allows specification of an optional reduction operation, which is applied to all values in `updates`
        tensor into `output` at the specified `indices`.
        In cases where `reduction` is set to "none", indices should not have duplicate entries: that is, if idx1 != idx2,
        then indices[idx1] != indices[idx2]. For instance, in a 2-D tensor case, the update
        corresponding to the [i][j] entry is performed as below:
        ::

              output[indices[i][j]][j] = updates[i][j] if axis = 0,
              output[i][indices[i][j]] = updates[i][j] if axis = 1,


        When `reduction` is set to "add", the update corresponding to the [i][j] entry is performed as below:
        ::

              output[indices[i][j]][j] += updates[i][j] if axis = 0,
              output[i][indices[i][j]] += updates[i][j] if axis = 1,


        When `reduction` is set to "mul", the update corresponding to the [i][j] entry is performed as below:
        ::

              output[indices[i][j]][j] *= updates[i][j] if axis = 0,
              output[i][indices[i][j]] *= updates[i][j] if axis = 1,


        This operator is the inverse of GatherElements. It is similar to Torch's Scatter operation.
        Example 1:
        ::

              data = [
                  [0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0],
              ]
              indices = [
                  [1, 0, 2],
                  [0, 2, 1],
              ]
              updates = [
                  [1.0, 1.1, 1.2],
                  [2.0, 2.1, 2.2],
              ]
              output = [
                  [2.0, 1.1, 0.0]
                  [1.0, 0.0, 2.2]
                  [0.0, 2.1, 1.2]
              ]


        Example 2:
        ::

              data = [[1.0, 2.0, 3.0, 4.0, 5.0]]
              indices = [[1, 3]]
              updates = [[1.1, 2.1]]
              axis = 1
              output = [[1.0, 1.1, 3.0, 2.1, 5.0]]




        Args:
            data: (differentiable) Tensor of rank r >= 1.

            indices: (non-differentiable) Tensor of int32/int64 indices, of r >= 1 (same
                rank as input). All index values are expected to be within bounds [-s,
                s-1] along axis of size s. It is an error if any of the index values are
                out of bounds.

            updates: (differentiable) Tensor of rank r >=1 (same rank and shape as
                indices)

            axis: Which axis to scatter on. Negative value means counting dimensions
                from the back. Accepted range is [-r, r-1] where r = rank(data).

            reduction: Type of reduction to apply: none (default), add, mul. 'none': no
                reduction applied. 'add':  reduction using the addition operation.
                'mul': reduction using the multiplication operation.
        """

        schema = get_schema("ScatterElements", 16, "")
        op = Op(self, "ScatterElements", schema)
        return op(
            *self._prepare_inputs(schema, data, indices, updates),
            axis=axis,
            reduction=reduction,
        )

    T_ScatterND = TypeVar(
        "T_ScatterND",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def ScatterND(
        self,
        data: T_ScatterND,
        indices: INT64,
        updates: T_ScatterND,
        *,
        reduction: str = "none",
    ) -> T_ScatterND:
        r"""[🌐 ScatterND(16)](https://onnx.ai/onnx/operators/onnx__ScatterND.html#scatternd-16 "Online Documentation")


        ScatterND takes three inputs `data` tensor of rank r >= 1, `indices` tensor of rank q >= 1,
        and `updates` tensor of rank q + r - indices.shape[-1] - 1. The output of the operation
        is produced by creating a copy of the input `data`, and then updating its value to values
        specified by `updates` at specific index positions specified by `indices`. Its output shape
        is the same as the shape of `data`.

        `indices` is an integer tensor. Let k denote indices.shape[-1], the last dimension in the shape of `indices`.
         `indices` is treated as a (q-1)-dimensional tensor of k-tuples, where each k-tuple is a partial-index into `data`.
        Hence, k can be a value at most the rank of `data`. When k equals rank(data), each update entry specifies an
        update to a single element of the tensor. When k is less than rank(data) each update entry specifies an
        update to a slice of the tensor. Index values are allowed to be negative, as per the usual
        convention for counting backwards from the end, but are expected in the valid range.

        `updates` is treated as a (q-1)-dimensional tensor of replacement-slice-values. Thus, the
        first (q-1) dimensions of updates.shape must match the first (q-1) dimensions of indices.shape.
        The remaining dimensions of `updates` correspond to the dimensions of the
        replacement-slice-values. Each replacement-slice-value is a (r-k) dimensional tensor,
        corresponding to the trailing (r-k) dimensions of `data`.  Thus, the shape of `updates`
        must equal indices.shape[0:q-1] ++ data.shape[k:r-1], where ++ denotes the concatenation
        of shapes.

        The `output` is calculated via the following equation:
            output = np.copy(data)
            update_indices = indices.shape[:-1]
            for idx in np.ndindex(update_indices):
                output[indices[idx]] = updates[idx]
        The order of iteration in the above loop is not specified.
        In particular, indices should not have duplicate entries: that is, if idx1 != idx2, then indices[idx1] != indices[idx2].
        This ensures that the output value does not depend on the iteration order.

        `reduction` allows specification of an optional reduction operation, which is applied to all values in `updates`
        tensor into `output` at the specified `indices`.
        In cases where `reduction` is set to "none", indices should not have duplicate entries: that is, if idx1 != idx2,
        then indices[idx1] != indices[idx2]. This ensures that the output value does not depend on the iteration order.
        When `reduction` is set to "add", `output` is calculated as follows:
            output = np.copy(data)
            update_indices = indices.shape[:-1]
            for idx in np.ndindex(update_indices):
                output[indices[idx]] += updates[idx]
        When `reduction` is set to "mul", `output` is calculated as follows:
            output = np.copy(data)
            update_indices = indices.shape[:-1]
            for idx in np.ndindex(update_indices):
                output[indices[idx]] *= updates[idx]
        This operator is the inverse of GatherND.
        Example 1:
        ::

              data    = [1, 2, 3, 4, 5, 6, 7, 8]
              indices = [[4], [3], [1], [7]]
              updates = [9, 10, 11, 12]
              output  = [1, 11, 3, 10, 9, 6, 7, 12]


        Example 2:
        ::

              data    = [[[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
                         [[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
                         [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]],
                         [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]]]
              indices = [[0], [2]]
              updates = [[[5, 5, 5, 5], [6, 6, 6, 6], [7, 7, 7, 7], [8, 8, 8, 8]],
                         [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]]]
              output  = [[[5, 5, 5, 5], [6, 6, 6, 6], [7, 7, 7, 7], [8, 8, 8, 8]],
                         [[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
                         [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]],
                         [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]]]




        Args:
            data: (differentiable) Tensor of rank r >= 1.

            indices: (non-differentiable) Tensor of rank q >= 1.

            updates: (differentiable) Tensor of rank q + r - indices_shape[-1] - 1.

            reduction: Type of reduction to apply: none (default), add, mul. 'none': no
                reduction applied. 'add':  reduction using the addition operation.
                'mul': reduction using the multiplication operation.
        """

        schema = get_schema("ScatterND", 16, "")
        op = Op(self, "ScatterND", schema)
        return op(*self._prepare_inputs(schema, data, indices, updates), reduction=reduction)

    B_Where: TypeAlias = BOOL

    T_Where = TypeVar(
        "T_Where",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Where(self, condition: B_Where, X: T_Where, Y: T_Where) -> T_Where:
        r"""[🌐 Where(16)](https://onnx.ai/onnx/operators/onnx__Where.html#where-16 "Online Documentation")


        Return elements, either from X or Y, depending on condition.
        Where behaves like
        [numpy.where](https://docs.scipy.org/doc/numpy/reference/generated/numpy.where.html)
        with three parameters.

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        Args:
            condition: (non-differentiable) When True (nonzero), yield X, otherwise
                yield Y

            X: (differentiable) values selected at indices where condition is True

            Y: (differentiable) values selected at indices where condition is False
        """

        schema = get_schema("Where", 16, "")
        op = Op(self, "Where", schema)
        return op(*self._prepare_inputs(schema, condition, X, Y))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Does the following transformation:
- Add(MatMul(X, W), B) -> Gemm
- Add(MatMul(Transpose(X), W), B) -> Gemm
- Add(MatMul(X, Transpose(W)), B) -> Gemm
- Add(MatMul(Transpose(X), Transpose(W)), B) -> Gemm
"""

import abc
from typing import ClassVar

from onnxscript.rewriter import _ir_utils
from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._rewrite_rule import RewriteRuleClassBase, RewriteRuleSet


class _MatMulAddToGemmBase(RewriteRuleClassBase, abc.ABC):
    trans_a: ClassVar = False
    trans_b: ClassVar = False

    def rewrite(self, op, input_a, input_b, input_c):
        attributes = {}
        if self.trans_a:
            attributes["transA"] = 1
        if self.trans_b:
            attributes["transB"] = 1
        return op.Gemm(input_a, input_b, input_c, **attributes)

    def check(self, context, input_a, input_b, **_):
        del context  # Not used
        check_result = MatchResult()
        # Rank of input_a and input_b must be 2
        if not (_ir_utils.has_rank(input_a, 2) and _ir_utils.has_rank(input_b, 2)):
            return check_result.fail("Rank of input_a and input_b must be 2")
        return check_result


class MatMulAddToGemm(_MatMulAddToGemmBase):
    """Replaces ``Add(MatMul(a, b), c)`` with ``Gemm(a, b, c)``."""

    def pattern(self, op, input_a, input_b, input_c):
        matmul = op.MatMul(input_a, input_b)
        return op.Add(matmul, input_c)


class TransAMatMulAddToGemm(_MatMulAddToGemmBase):
    """Replaces ``Add(MatMul(Transpose(a), b), c)`` with ``Gemm(a, b, c)``."""

    trans_a: ClassVar = True

    def pattern(self, op, input_a, input_b, input_c):
        matmul = op.MatMul(op.Transpose(input_a, perm=[1, 0]), input_b)
        return op.Add(matmul, input_c)


class TransBMatMulAddToGemm(_MatMulAddToGemmBase):
    """Replaces ``Add(MatMul(a, Transpose(b)), c)`` with ``Gemm(a, b, c)``."""

    trans_b: ClassVar = True

    def pattern(self, op, input_a, input_b, input_c):
        matmul = op.MatMul(input_a, op.Transpose(input_b, perm=[1, 0]))
        return op.Add(matmul, input_c)


class TransABMatMulAddToGemm(_MatMulAddToGemmBase):
    """Replaces ``Add(MatMul(Transpose(a), Transpose(b)), c)`` with ``Gemm(a, b, c)``."""

    trans_a: ClassVar = True
    trans_b: ClassVar = True

    def pattern(self, op, input_a, input_b, input_c):
        matmul = op.MatMul(
            op.Transpose(input_a, perm=[1, 0]),
            op.Transpose(input_b, perm=[1, 0]),
        )
        return op.Add(matmul, input_c)


matmul_add_to_gemm_rule = MatMulAddToGemm().rule()
transpose_a_matmul_add_to_gemm_rule = TransAMatMulAddToGemm().rule()
transpose_b_matmul_add_to_gemm_rule = TransBMatMulAddToGemm().rule()
transpose_ab_matmul_add_to_gemm_rule = TransABMatMulAddToGemm().rule()


rules = RewriteRuleSet(
    [
        transpose_ab_matmul_add_to_gemm_rule,
        transpose_a_matmul_add_to_gemm_rule,
        transpose_b_matmul_add_to_gemm_rule,
        matmul_add_to_gemm_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import logging

import numpy as np
import onnx

from onnxscript.rewriter._rewrite_rule import RewriteRule, RewriteRuleSet

logger = logging.getLogger(__name__)


def check_if_simulated_instance_norm_is_used(
    context,
    input_x,
    adjusted_input_shape,
    original_input_shape,
    weight_for_norm,
    bias_for_norm,
    weight_full,
    bias_full,
    **_,
) -> bool:
    """Check if the simulated instance normalization is used.

    In torchlib with opset18, onnx.GroupNorm is using wrong definition, so
    we use InstanceNormalization to simulate GroupNormalization. We need to check if there are arguments created to simulation.
    If there are, then we need to replace the pattern. If they are not used, then we don't need to replace the pattern.

    To validate this, we need to check the following:
    1. weight_for_norm are all 1 and bias_for_norm are all 0, as they are created for the simulation.
    2. weight_full and bias_full are unsqueezed to be easily broadcastable.
    3. input rank should be 4
    4. weight_full and bias_full should have ones except first dim.
    5. adjusted_input_shape is a constant tensor of form [0, g, -1]
    6. original_input_shape is the same as input_x shape.

    Returns:
        bool: True if the simulated instance normalization is used, False otherwise.
    """
    weight_for_norm_const_value = weight_for_norm.const_value
    if weight_for_norm_const_value is None:
        return False
    weight_for_norm = weight_for_norm_const_value.numpy()

    bias_for_norm_const_value = bias_for_norm.const_value
    if bias_for_norm_const_value is None:
        return False
    bias_for_norm = bias_for_norm_const_value.numpy()

    if not np.all(weight_for_norm == 1):
        return False
    if not np.all(bias_for_norm == 0):
        return False

    input_rank_minus_one = len(input_x.shape) - 1
    weight_full_rank = len(weight_full.shape)
    bias_full_rank = len(bias_full.shape)
    if weight_full_rank != input_rank_minus_one or bias_full_rank != input_rank_minus_one:
        return False

    input_rank = len(input_x.shape)
    if input_rank != 4:
        return False

    weight_full_shape = weight_full.shape
    if not all(dim == 1 for dim in weight_full_shape[1:]):
        return False
    bias_full_shape = bias_full.shape
    if not all(dim == 1 for dim in bias_full_shape[1:]):
        return False

    adjusted_input_shape_const_value = adjusted_input_shape.const_value

    g = weight_for_norm.shape[0]
    if (
        adjusted_input_shape_const_value is None
        or adjusted_input_shape_const_value.numpy().tolist() != [0, g, -1]
    ):
        return False

    # NOTE: Restrict the rule to only support constant shape
    original_input_shape_const_value = original_input_shape.const_value
    if (
        original_input_shape_const_value is None
        or original_input_shape_const_value.numpy().tolist() != input_x.shape
    ):
        return False

    return True


def instance_simulates_group_normalization_pattern(
    op,
    input_x,
    adjusted_input_shape,
    original_input_shape,
    weight_for_norm,
    bias_for_norm,
    weight_full,
    bias_full,
    epsilon,
):
    adjusted_input = op.Reshape(input_x, adjusted_input_shape)
    inst_norm = op.InstanceNormalization(
        adjusted_input, weight_for_norm, bias_for_norm, epsilon=epsilon
    )
    adjusted_inst_norm = op.Reshape(inst_norm, original_input_shape)
    mul = op.Mul(adjusted_inst_norm, weight_full)
    return op.Add(mul, bias_full)


def group_normalization(op, input_x, weight_for_norm, weight_full, bias_full, epsilon, **_):
    # com.microsoft.GroupNorm only supports NHWC for now
    nhwc_input = op.Transpose(input_x, perm=[0, 2, 3, 1])
    # com.microsoft.GroupNorm only supports gamma and beta as float type
    weight_full = op.Cast(weight_full, to=onnx.TensorProto.FLOAT)
    reshape_to_1d = op.Constant(value_ints=[-1])
    weight_full = op.Reshape(weight_full, reshape_to_1d)
    bias_full = op.Cast(bias_full, to=onnx.TensorProto.FLOAT)
    bias_full = op.Reshape(bias_full, reshape_to_1d)
    # re-obtain attribute groups
    # TODO(rama): Earlier check implies weight_for_norm is a constant tensor?
    # If not, we should add a check that shape[0] is not symbolic.
    shape = weight_for_norm.shape
    if shape is None:
        raise ValueError("weight_for_norm shape not known")
    groups = shape[0]
    output = op.GroupNorm(
        nhwc_input,
        weight_full,
        bias_full,
        activation=0,
        channels_last=1,
        epsilon=epsilon,
        groups=groups,
        _domain="com.microsoft",
    )
    return op.Transpose(output, perm=[0, 3, 1, 2])


# Register the rewrite rules
instance_norm_to_group_norm_rule = RewriteRule(
    instance_simulates_group_normalization_pattern,
    group_normalization,
    check_if_simulated_instance_norm_is_used,
)

# NOTE: instance_norm_to_group_norm_rule is subset of instance_norm_to_group_norm_with_silu_rule,
# so we need to run instance_norm_to_group_norm_with_silu_rule first.
rules = RewriteRuleSet([instance_norm_to_group_norm_rule])



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

__all__ = [
    "TorchScriptTensor",
    "TorchScriptGraph",
    "TorchScriptTracingEvaluator",
]


class _RemovedClass:
    """A onnxscript tensor that wraps a torchscript Value."""

    def __init__(self, *_, **__):
        raise NotImplementedError(
            "Support for dynamo_export has been dropped since onnxscript 0.4.0. "
            "Please use `torch.onnx.export(..., dynamo=True)`, or downgrade to onnxscript<0.4"
        )


TorchScriptTensor = _RemovedClass
TorchScriptGraph = _RemovedClass
TorchScriptTracingEvaluator = _RemovedClass
